"""
Filtering functions for protein structural classification by SCOP.

These all are applied to individual atom dataframes, and remove entries from that dataframe as necessary.
"""
import numpy as np
import pandas as pd

import atom3d.protein.scop as scop
import atom3d.util.file as fi


def form_scop_filter(level, allowed=None, excluded=None):
    """
    Create filter by SCOP classification at a specified level.

    Valid levels are: type, class, fold, superfamily, family.

    :param level: values that we will keep when they are found.
    :type level: str
    :param allowed: allowed SCOP values.
    :type allowed: list[str], optional
    :param excluded: excluded SCOP values.
    :type excluded: list[str], optional

    :return: function that implements the specified filter.
    :rtype: filter function
    """
    if excluded is not None and allowed is not None:
        raise RuntimeError('Can only specify one of allowed and excluded.')
    if allowed is None:
        allowed = []
    if excluded is None:
        excluded = []
    scop_index = scop.get_scop_index()
    scop_index = scop_index[level]

    # Build quick lookup tables.
    if allowed:
        permitted = pd.Series(
            {pdb_code: x.drop_duplicates().isin(allowed).any()
             for pdb_code, x in scop_index.groupby('pdb_code')})
    if excluded:
        forbidden = pd.Series(
            {pdb_code: x.drop_duplicates().isin(excluded).any()
             for pdb_code, x in scop_index.groupby('pdb_code')})

    def filter_fn(df):
        pdb_codes = df['structure'].apply(lambda x: x[:4].lower())

        if len(allowed) > 0:
            to_keep = permitted[pdb_codes]
            # If didn't find, be conservative and do not use.
            to_keep[to_keep.isna()] = False
            to_keep = to_keep.astype(bool)
        elif len(excluded) > 0:
            to_exclude = forbidden[pdb_codes]
            # If didn't find, be conservative and do not use.
            to_exclude[to_exclude.isna()] = True
            to_exclude = to_exclude.astype(bool)
            to_keep = ~to_exclude
        else:
            to_keep = pd.Series([True] * len(df), index=df['structure'])
        return df[to_keep.values]
    return filter_fn


def form_scop_filter_against(dataset, level, conservative):
    """
    Create filter that removes structures with matching SCOP class to a chain in supplied dataset.

    We consider each chain in each structure separately, and remove the structure if any of them matches any chain in dataset. This is done at the specified SCOP level.  Valid levels are: type, class, fold, superfamily, family.

    :param dataset: dataset that if we are checking for matches against.
    :type dataset: atom3d dataset
    :param level: SCOP level at which we are comparing datasets.
    :type level: str
    :param conservative: indicates what we should do about pdbs that do not have any SCOP class associated with them.  True means we throw out, False means we keep.
    :type conservative: bool

    :return: function that implements the specified filter.
    :rtype: filter function.
    """
    scop_index = scop.get_scop_index()[level]

    def form_scop_against():
        result = []
        for x in dataset:
            for (e, su, st), structure in x['atoms'].groupby(
                    ['ensemble', 'subunit', 'structure']):
                pc = fi.get_pdb_code(st).lower()
                for (m, c), _ in structure.groupby(['model', 'chain']):
                    if (pc, c) in scop_index:
                        result.append(scop_index.loc[(pc, c)].values)
        return np.unique(np.concatenate(result))
    scop_against = form_scop_against()

    def filter_fn(df):
        to_keep = {}
        for (e, su, st), structure in df.groupby(
                ['ensemble', 'subunit', 'structure']):
            pc = fi.get_pdb_code(st).lower()
            for (m, c), _ in structure.groupby(['model', 'chain']):
                if (pc, c) in scop_index:
                    scop_found = scop_index.loc[(pc, c)].values
                    if np.isin(scop_found, scop_against).any():
                        to_keep[(st, m, c)] = False
                    else:
                        to_keep[(st, m, c)] = True
                else:
                    to_keep[(st, m, c)] = not conservative
        to_keep = \
            pd.Series(to_keep)[pd.Index(df[['structure', 'model', 'chain']])]
        return df[to_keep.values]
    return filter_fn


import io
import os
from pathlib import Path
import subprocess

import pandas as pd
import tqdm

import atom3d.util.file as fi


class Scores(object):
    """
    Class for tracking and looking up Rosetta score files.

    :param file_list: List of paths to Rosetta silent files.
    :type file_list: list[Union[str, Path]]
    """

    def __init__(self, file_list):
        self._scores = {}
        file_list = [Path(x).absolute() for x in file_list]
        for silent_file in file_list:
            key = self._key_from_silent_file(silent_file)
            self._scores[key] = self._parse_scores(silent_file)

        self._scores = pd.concat(self._scores).sort_index()

    def _parse_scores(self, silent_file):
        grep_cmd = f"grep ^SCORE: {silent_file}"
        out = subprocess.Popen(
            grep_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            cwd=os.getcwd(), shell=True)
        (stdout, stderr) = out.communicate()

        f = io.StringIO(stdout.decode('utf-8'))
        return pd.read_csv(f, delimiter='\s+').drop('SCORE:', axis=1) \
            .set_index('description')

    def _key_from_silent_file(self, silent_file):
        return silent_file.stem.split('.')[0]

    def _lookup(self, file_path):
        file_path = Path(file_path)
        key = (file_path.stem, file_path.name)
        if key in self._scores.index:
            return key, self._scores.loc[key]
        key = (file_path.parent.stem, file_path.stem)
        if key in self._scores.index:
            return key, self._scores.loc[key]
        return None, None

    def __call__(self, x, error_if_missing=False):
        key, x['scores'] = self._lookup(x['file_path'])
        if key is not None:
            x['id'] = str(key)
        if x['scores'] is None and error_if_missing:
            raise RuntimeError(f'Unable to find scores for {x["file_path"]}')
        return x

    def remove_missing(self, file_list):
        """Remove examples we cannot find in score files."""
        result = []
        for i, file_path in tqdm.tqdm(enumerate(file_list), total=len(file_list)):
            entry = self._lookup(file_path)
            if entry is not None:
                result.append(file_path)
        return result

    def __len__(self):
        """Get number of individual structures across score files."""
        return len(self._scores)



import logging
import sys


def get_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    if not logger.handlers:
        handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter(
            '%(asctime)s %(levelname)s %(process)d: %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    logger.propagate = False
    return logger


"""Code relating to biomolecular sequence."""
import os
import subprocess

import Bio.PDB.Polypeptide as Poly
import dotenv as de
import numpy as np
from Bio.Blast.Applications import NcbiblastpCommandline

import atom3d.util.log as log

project_root = os.path.abspath(os.path.join(__file__, '../../..'))
de.load_dotenv(os.path.join(project_root, '.env'))
logger = log.get_logger('sequence')


def find_similar(chain_sequences, blast_db, cutoff, num_alignments):
    """Find all other proteins that have sequence identity greater than ``cutoff`` to query protein given by ``chain_sequences``. Assumes sequences have already been processed into a BLAST-formatted database by :func:`write_to_blast_db`.

    :param chain_sequences: Query protein represented as a list of tuples of (id, sequence) for each chain, as output by :func:`get_chain_sequences`.
    :type chain_sequences: Tuple[Tuple[str, str, str, str], str]
    :param blast_db: Location of BLAST database
    :type blast_db: str
    :param cutoff: Sequence identity cutoff; must be between 0 and 1
    :type cutoff: float
    :param num_alignments: Number of proteins for which to calculate identity.
    :type num_alignments: int
    
    :return: List of chain identifiers (as tuples) for proteins found within specified sequence identity.
    :rtype: List[Tuple[str, str, str, str]]
    """    
    """Find all other pdbs that have sequence identity greater than cutoff."""
    
    if 'BLAST_BIN' not in os.environ:
        raise RuntimeError('Need to set BLAST_BIN in .env to use blastp')
    if not (0 <= cutoff <= 1):
        raise Exception('cutoff need to be between 0 and 1')

    sim = set()
    for chain, s in chain_sequences:
        blastp_cline = NcbiblastpCommandline(
            db=blast_db,
            outfmt="10 nident sacc",
            num_alignments=num_alignments,
            cmd=os.path.join(os.environ['BLAST_BIN'], 'blastp'))
        out, err = blastp_cline(stdin=s)

        for res in out.split():
            nident, match = res.split(',')
            seq_id = float(nident)*1.0 / len(s)
            if seq_id >= cutoff:
                sim.add(_fasta_name_to_tuple(match))
    return list(sim)


####################################
# Sequence-level clustering code
####################################


def get_pdb_clusters(id_level, pdb_ids=None):
    """
    Downloads pre-calculated clusters from PDB at given cutoff.
    Returns dictionaries mapping PDB IDs to cluster IDs and vice versa.
    """
    id_level = int(id_level * 100)
    if id_level not in [30, 40, 50, 70, 90, 95, 100]:
        raise Exception(
            'invalid invalid identity cutoff. '
            'possible values = 30,40,50,70,90,95,100')
    logger.info('getting clusters from PDB...')
    if not os.path.exists(f'bc-{id_level}.out'):
        subprocess.call(
            'wget ftp://resources.rcsb.org/sequence/clusters/bc-{}.out'.format(
                id_level),
            shell=True)
    pdb2cluster = {}
    cluster2pdb = {}
    with open(f'bc-{id_level}.out') as f:
        for i, line in enumerate(f):
            cluster2pdb[i] = set()
            pdbs = line.strip().split()
            pdbs = [p[:4].lower() for p in pdbs]
            for pdb in pdbs:
                if pdb_ids is not None and not np.isin(pdb, pdb_ids).any():
                    continue
                if pdb not in pdb2cluster:
                    pdb2cluster[pdb] = set()
                pdb2cluster[pdb].add(i)
                cluster2pdb[i].add(pdb)

    return pdb2cluster, cluster2pdb


def find_cluster_members(pdb, clusterings):
    """
    Find all other pdbs that co-occur in a cluster with a given pdb
    """
    pdb2cluster, cluster2pdb = clusterings
    clusters = pdb2cluster[pdb]
    all_pdbs = set()
    for clust in clusters:
        pdbs = cluster2pdb[clust]
        all_pdbs = all_pdbs.union(pdbs)
    return all_pdbs


####################################
# useful functions
####################################


def write_to_blast_db(chain_sequences, blast_db):
    """Write provided sequences to blast db, for use with BLAST. Sequences should be provided as (id, sequence) tuples as returned by :func:`get_chain_sequences`.
    """
    logger.info(f'writing {len(chain_sequences):} entries to BLAST db {blast_db:}')

    if 'BLAST_BIN' not in os.environ:
        raise RuntimeError('Need to set BLAST_BIN in .env to use makeblastdb')

    if os.path.exists(blast_db):
        logger.warning(f'Removing previous blast db at {blast_db:}...')
        os.remove(blast_db)
    _write_fasta(chain_sequences, blast_db)
    cmd = os.path.join(os.environ['BLAST_BIN'], 'makeblastdb')
    subprocess.check_output(f'{cmd:} -in {blast_db:} -dbtype prot', shell=True)


def _tuple_to_fasta_name(name_tuple):
    """Fasta names can only be strings. This code writes the tuples in."""
    return '___'.join([str(x) for x in name_tuple])


def _fasta_name_to_tuple(x):
    """Fasta names can only be strings. This code gets the tuples back out."""
    return tuple(x.split('___'))


def get_chain_sequences(df):
    """Return list of tuples of (id, sequence) for different chains of monomers in a given dataframe."""
    # Keep only CA of standard residues
    df = df[df['name'] == 'CA'].drop_duplicates()
    df = df[df['resname'].apply(lambda x: Poly.is_aa(x, standard=True))]
    df['resname'] = df['resname'].apply(Poly.three_to_one)
    chain_sequences = []
    for c, chain in df.groupby(['ensemble', 'subunit', 'structure', 'model', 'chain']):
        seq = ''.join(chain['resname'])
        chain_sequences.append((tuple([str(x) for x in c]), seq))
    return chain_sequences


def _write_fasta(chain_sequences, outfile):
    """
    Write chain_sequences to fasta file. 
    """
    with open(outfile, 'w') as f:
        for chain, seq in chain_sequences:
            f.write('>' + _tuple_to_fasta_name(chain) + '\n')
            f.write(seq + '\n')


"""Code for sharding structures."""
import logging
import multiprocessing as mp
import os
import shutil

import click
import numpy as np
import pandas as pd
import tqdm

import atom3d.shard.ensemble as en
import atom3d.util.file as fi
import atom3d.util.formats as dt

db_sem = mp.Semaphore()


@click.command(help='Combine files into sharded HDF5 files.')
@click.argument('input_dir', type=click.Path(exists=True))
@click.argument('sharded_path')
@click.option('--filetype', type=click.Choice(['pdb', 'pdb.gz', 'mmcif', 'sdf']),
              default='pdb', help='which kinds of files are we sharding.')
@click.option('--ensembler', type=click.Choice(en.ensemblers.keys()),
              default='none', help='how to ensemble files')
def shard_dataset(input_dir, sharded_path, filetype, ensembler):
    """Shard whole input dataset."""
    logging.basicConfig(format='%(asctime)s %(levelname)s %(process)d: ' +
                        '%(message)s',
                        level=logging.INFO)

    dirname = os.path.dirname(sharded_path)
    if not os.path.exists(dirname) and dirname != '':
        os.makedirs(dirname, exist_ok=True)

    files = fi.find_files(input_dir, dt.patterns[filetype])
    ensemble_map = en.ensemblers[ensembler](files)
    Sharded.create_from_ensemble_map(ensemble_map, sharded_path)


class Sharded(object):
    """Sharded pandas dataframe representation."""

    def __init__(self, path, keys):
        self.path = path
        self._keys = keys

    @classmethod
    def create_from_ensemble_map(cls, ensemble_map, path):
        sharded = cls(path, ['ensemble'])

        num_shards = sharded.get_num_shards()

        # Check if already partly written.  If so, resume from there.
        metadata_path = sharded._get_metadata()
        if os.path.exists(metadata_path):
            metadata = pd.read_hdf(metadata_path, f'atom3d/data/metadata')
            num_written = len(metadata['shard_num'].unique())
        else:
            num_written = 0

        shard_ranges = _get_shard_ranges(len(ensemble_map), num_shards)
        shard_size = shard_ranges[0, 1] - shard_ranges[0, 0]

        logging.info(f'Ensembles per shard: {shard_size:}')
        for shard_num in tqdm.trange(num_written, num_shards):
            start, stop = shard_ranges[shard_num]

            dfs = []
            for name in sorted(ensemble_map.keys())[start:stop]:
                df = en.parse_ensemble(name, ensemble_map[name])
                dfs.append(df)
            df = dt.merge_dfs(dfs)

            sharded._write_shard(shard_num, df)

    @classmethod
    def load(cls, path):
        """Load a fully written sharded dataset."""

        # Get keys from metadata file.
        sharded = cls(path, None)
        metadata_path = sharded._get_metadata()
        if not os.path.exists(metadata_path):
            raise RuntimeError(f'Metadata for {path:} does not exist')
        metadata = pd.read_hdf(metadata_path, f'metadata')
        keys = metadata.columns.tolist()
        keys.remove('shard_num')
        keys.remove('start')
        keys.remove('stop')

        sharded = cls(path, keys)
        if not sharded.is_written():
            raise RuntimeError(
                f'Sharded loaded from {path:} not fully written.')
        return sharded

    def is_written(self):
        """Check if metadata files and all data files are there."""
        metadata_path = self._get_metadata()
        if not os.path.exists(metadata_path):
            return False
        num_shards = self.get_num_shards()
        for i in range(num_shards):
            shard = self._get_shard(i)
            if not os.path.exists(shard):
                return False
        return True

    def get_keys(self):
        return self._keys

    def iter_shards(self):
        """Iterate through shards."""
        num_shards = self.get_num_shards()
        for i in range(num_shards):
            yield i, self.read_shard(i)

    def read_shard(self, shard_num, key='structures'):
        """Read a single shard of a sharded dataset."""
        shard = self._get_shard(shard_num)
        return pd.read_hdf(shard, key)

    def read_keyed(self, name):
        """Read keyed entry from sharded dataset."""
        metadata_path = self._get_metadata()
        metadata = pd.read_hdf(metadata_path, f'metadata')
        entry = metadata[(metadata[self._keys] == name).any(axis=1)]
        if len(entry) != 1:
            raise RuntimeError('Need exactly one matchin in structure lookup')
        entry = entry.iloc[0]

        shard = self._get_shard(entry['shard_num'])
        df = pd.read_hdf(shard, 'structures',
                         start=entry['start'], stop=entry['stop'])
        return df.reset_index(drop=True)

    def get_names(self):
        """Get keyed names in sharded dataset."""
        metadata_path = self._get_metadata()
        metadata = pd.read_hdf(metadata_path, f'metadata')
        return metadata[self._keys]

    def get_num_keyed(self):
        """Get number of keyed examples in sharded dataset."""
        return self.get_names().shape[0]

    def get_num_structures(self, keys):
        """Get number of structures in sharded dataset."""
        num_structs = 0
        for _, df in self.iter_shards():
            num_structs += df.groupby(keys).ngroups
        return num_structs

    def move(self, dest_path):
        """Move sharded dataset."""
        self.copy(dest_path)
        self.delete_files()
        self.path = dest_path

    def copy(self, dest_path):
        """Copy sharded dataset."""
        num_shards = self.get_num_shards()

        dest_sharded = Sharded(dest_path, self._keys)

        for i in range(num_shards):
            source_shard = self._get_shard(i)
            dest_shard = dest_sharded._get_shard(i)
            if os.path.exists(source_shard):
                shutil.copyfile(source_shard, dest_shard)
        metadata_path = self._get_metadata()
        dest_metadata_path = dest_sharded._get_metadata()
        if os.path.exists(metadata_path):
            shutil.copyfile(metadata_path, dest_metadata_path)
        return dest_sharded

    def delete_files(self):
        """Delete sharded dataset."""
        num_shards = self.get_num_shards()
        for i in range(num_shards):
            shard = self._get_shard(i)
            if os.path.exists(shard):
                os.remove(shard)
        metadata_path = self._get_metadata()
        if os.path.exists(metadata_path):
            os.remove(metadata_path)

    def add_to_shard(self, shard_num, df, key):
        """Add dataframe under key to shard."""
        shard = self._get_shard(shard_num)
        df.to_hdf(shard, key, mode='a')

    def has(self, shard_num, key):
        """If key is present in shard."""
        shard = self._get_shard(shard_num)
        with pd.HDFStore(shard, mode='r') as f:
            return key in [x[1:] for x in f.keys()]

    def get_num_shards(self):
        """Get number of shards in sharded dataset."""
        return get_num_shards(self.path)

    def get_prefix(self):
        return '@'.join(self.path.split('@')[:-1])

    def _get_shard(self, shard_num):
        num_shards = self.get_num_shards()
        prefix = get_prefix(self.path)
        return f'{prefix:}_{shard_num:04d}_{num_shards:}.h5'

    def _get_metadata(self):
        num_shards = self.get_num_shards()
        prefix = get_prefix(self.path)
        return f'{prefix:}_meta_{num_shards:}.h5'

    def _write_shard(self, shard_num, df):
        """Write to a single shard of a sharded dataset."""

        if len(self._keys) == 1:
            metadata = pd.DataFrame(
                [(shard_num, y.index[0], y.index[0] + len(y)) + (x,)
                 for x, y in dt.split_df(df, self._keys)],
                columns=['shard_num', 'start', 'stop'] + self._keys)
        else:
            metadata = pd.DataFrame(
                [(shard_num, y.index[0], y.index[0] + len(y)) + x
                 for x, y in dt.split_df(df, self._keys)],
                columns=['shard_num', 'start', 'stop'] + self._keys)

        path = self._get_shard(shard_num)
        df.to_hdf(path, f'structures')
        with db_sem:
            # Check that we are writing same name again to same sharded
            # dataset.
            metadata_path = self._get_metadata()
            if os.path.exists(metadata_path):
                metadata = pd.concat((pd.read_hdf(metadata_path, f'metadata'),
                                      metadata)).reset_index(drop=True)
                if metadata[self._keys].duplicated().any():
                    raise RuntimeError(f'Writing duplicate to sharded {path:}')

            metadata.to_hdf(metadata_path, f'metadata', mode='w')


def get_prefix(path):
    return '@'.join(path.split('@')[:-1])


def get_num_shards(path):
    """Get number of shards in sharded dataset."""
    return int(path.split('@')[-1])


def _get_shard_ranges(num_structures, num_shards):
    """Get list of shard starts and ends."""
    base_shard_size = int(num_structures / num_shards)
    excess = num_structures - base_shard_size * num_shards
    shard_sizes = np.ones(num_shards, dtype=np.int32) * base_shard_size
    shard_sizes[:excess] += 1
    stops = np.cumsum(shard_sizes)
    starts = stops - shard_sizes
    return np.stack((starts, stops)).T


if __name__ == "__main__":
    shard_dataset()



"""TODO: This code has been significantly re-written and should be tested."""
import math
import random

import numpy as np

import atom3d.protein.sequence as seq
import atom3d.util.file as fi
import atom3d.util.log as log
import atom3d.splits.splits as splits

logger = log.get_logger('sequence_splits')

####################################
# split by pre-clustered sequence
# identity clusters from PDB
####################################


def cluster_split(dataset, cutoff, val_split=0.1,
                  test_split=0.1, min_fam_in_split=5, random_seed=None):
    """Splits pdb dataset using pre-computed sequence identity clusters from PDB, ensuring that no cluster spans multiple splits. 

    Clusters are selected randomly into validation and test sets, but to ensure that there is some diversity in each set (i.e. a split does not consist of a single sequence cluster), a minimum number of clusters in each split is enforced. . Some data examples may be removed in order to satisfy this constraint.
    
    This function assumes that the PDB code or PDB filename exists in the ``ensemble`` field of the ``atoms`` dataframe in the dataset.

    :param dataset: Dataset to perform the split on.
    :type dataset: ATOM3D Dataset
    :param cutoff: Sequence identity cutoff. Possible values: 0.3, 0.4, 0.5, 0.7, 0.9, 0.95, 1.0
    :type cutoff: float
    :param val_split: Fraction of data used in validation set, defaults to 0.1
    :type val_split: float, optional
    :param test_split: Fraction of data used in test set, defaults to 0.1
    :type test_split: float, optional
    :param min_fam_in_split: Minimum number of sequence clusters to be included in validation and test sets, defaults to 5
    :type min_fam_in_split: int, optional
    :param random_seed: Random seed for sampling clusters, defaults to None
    :type random_seed: int, optional

    :return: Tuple containing training, validation, and test sets, each as ATOM3D Dataset objects.
    :rtype: Tuple[Dataset]
    """    
    
    if random_seed is not None:
        np.random.seed(random_seed)

    logger.info('Loading chain sequences')
    all_chain_sequences = [seq.get_chain_sequences(x['atoms']) for x in dataset]

    pdb_codes = np.array([fi.get_pdb_code(x[0][0][0]) for x in all_chain_sequences])
    n_orig = len(np.unique(pdb_codes))
    clusterings = seq.get_pdb_clusters(cutoff, np.unique(pdb_codes))

    # If code not present in clustering, we don't use.
    to_use = [i for (i, x) in enumerate(pdb_codes) if x in clusterings[0]]
    n = len(np.unique(pdb_codes[to_use]))
    to_use = set(to_use)

    logger.info(
        f'Removing {n_orig - n:} / {n_orig:} '
        f'sequences due to not finding in clustering.')

    test_size = n * test_split
    val_size = n * val_split

    logger.info('generating validation set...')
    val_indices, to_use = _create_cluster_split(
        all_chain_sequences, clusterings, to_use, val_size, min_fam_in_split)
    logger.info('generating test set...')
    test_indices, to_use = _create_cluster_split(
        all_chain_sequences, clusterings, to_use, test_size, min_fam_in_split)
    train_indices = to_use

    return splits.split(dataset, train_indices, val_indices, test_indices)


def _create_cluster_split(all_chain_sequences, clusterings, to_use, split_size, min_fam_in_split):
    """
    Helper function for :func:`create_cluster_split`. Creates a single split of ``split_size`` elements while retaining diversity specified by ``min_fam_in_split``.
    Takes in ``all_chain_sequences`` and reference ``clusterings`` from PDB, as well as a list of valid indices to sample from, specified by ``to_use``.
    Returns indices of new split and indices remaining in dataset after removing those used in split.
    """
    dataset_size = len(all_chain_sequences)
    code_to_idx = {fi.get_pdb_code(y[0][0]): i for (i, x) in enumerate(all_chain_sequences) for y in x}

    all_indices = set(range(dataset_size))
    split, used = set(), all_indices.difference(to_use)
    while len(split) < split_size:
        i = random.sample(to_use, 1)[0]
        pdb_code = fi.get_pdb_code(all_chain_sequences[i][0][0][0])
        found = seq.find_cluster_members(pdb_code, clusterings)

        # Map back to source.
        found = set([code_to_idx[x] for x in found])
        found = found.difference(used)

        # ensure that at least min_fam_in_split families in each split
        max_fam_size = int(math.ceil(split_size / min_fam_in_split))
        split = split.union(list(found)[:max_fam_size])
        to_use = to_use.difference(found)
        used = used.union(found)

    return split, to_use


####################################
# split by calculating sequence identity
# to any example in training set
####################################


def identity_split(
        dataset, cutoff, val_split=0.1, test_split=0.1,
        min_fam_in_split=5, blast_db=None, random_seed=None):
    """Splits a dataset of proteins by sequence identity at specified cutoff value. Proteins are randomly selected to be placed in validation and test splits, along with all proteins within ``cutoff`` sequence identity (calculated by BLAST).

        To ensure that there is some diversity in each set (i.e. a split does not consist of a single sequence cluster), a minimum number of clusters in each split is enforced. Some data examples may be removed in order to satisfy this constraint.

        Note that the construction of this function means that it is effectively a cluster split. All examples within ``cutoff`` of the sampled query protein are added to validation set, meaning that some examples near the edge of the cluster may in fact share less than ``cutoff`` sequence identity with other proteins in the dataset.
        Therefore, this does not satisfy the constraints for a strict sequence identity cutoff: that 
        (1) no protein in validation split shares greater than ``cutoff`` sequence identity with any protein in the train set, and 
        (2) no protein in the test split shares greater than ``cutoff`` sequence identity with any protein in either train or validation sets.
        A function that satisfies this more strict definition is currently under development.



    :param dataset: Dataset to perform the split on.
    :type dataset: ATOM3D Dataset
    :param cutoff: Sequence identity cutoff, between 0 and 1
    :type cutoff: float
    :param val_split: Fraction of data used in validation set, defaults to 0.1
    :type val_split: float, optional
    :param test_split: Fraction of data used in test set, defaults to 0.1
    :type test_split: float, optional
    :param min_fam_in_split: Minimum number of sequence clusters to be included in validation and test sets, defaults to 5
    :type min_fam_in_split: int, optional
    :param random_seed: Random seed for sampling clusters, defaults to None
    :type random_seed: int, optional

    :return: Tuple containing training, validation, and test sets, each as ATOM3D Dataset objects.
    :rtype: Tuple[Dataset]
    """        

    all_chain_sequences = [seq.get_chain_sequences(x['atoms']) for x in dataset]
    # Flatten.
    flat_chain_sequences = [x for sublist in all_chain_sequences for x in sublist]

    # write all sequences to BLAST-formatted database
    if blast_db is None:
        seq.write_to_blast_db(flat_chain_sequences, 'blast_db')
        blast_db = 'blast_db'

    if random_seed is not None:
        np.random.seed(random_seed)

    n = len(dataset)
    test_size = n * test_split
    val_size = n * val_split

    to_use = set(range(len(all_chain_sequences)))
    logger.info('generating validation set...')
    val_indices, to_use = _create_identity_split(
        all_chain_sequences, cutoff, to_use, val_size, min_fam_in_split, blast_db)
    logger.info('generating test set...')
    test_indices, to_use = _create_identity_split(
        all_chain_sequences, cutoff, to_use, test_size, min_fam_in_split, blast_db)
    train_indices = to_use

    return splits.split(dataset, train_indices, val_indices, test_indices)


def _create_identity_split(all_chain_sequences, cutoff, to_use, split_size,
                           min_fam_in_split, blast_db):
    """
    Helper function for :func:`create_cluster_split`. Creates a single split of ``split_size`` elements while retaining diversity specified by ``min_fam_in_split``.
    Takes in ``all_chain_sequences`` and reference ``blast_db``, as well as a list of valid indices to sample from, specified by ``to_use``.
    Returns indices of new split and indices remaining in dataset after removing those used in split.
    """
    dataset_size = len(all_chain_sequences)
    chain_to_idx = {y[0]: i for (i, x) in enumerate(all_chain_sequences) for y in x}

    all_indices = set(range(dataset_size))
    split, used = set(), all_indices.difference(to_use)
    while len(split) < split_size:
        i = random.sample(to_use, 1)[0]

        # Get chains that match.
        found = seq.find_similar(all_chain_sequences[i], blast_db, cutoff, dataset_size)
        # Map back to source.
        found = set([chain_to_idx[x] for x in found])
        found = found.difference(used)

        # ensure that at least min_fam_in_split families in each split
        max_fam_size = int(math.ceil(split_size / min_fam_in_split))
        split = split.union(list(found)[:max_fam_size])
        to_use = to_use.difference(found)
        used = used.union(found)

    return split, to_use


import contextlib
import gzip
import importlib
import json
import io
import logging
import msgpack
from pathlib import Path
import pickle as pkl
import tqdm
import urllib.request
import subprocess

import Bio.PDB
import lmdb
import numpy as np
import pandas as pd
from torch.utils.data import Dataset, IterableDataset

import atom3d.util.rosetta as ar
import atom3d.util.file as fi
import atom3d.util.formats as fo

logger = logging.getLogger(__name__)


class LMDBDataset(Dataset):
    """
    Creates a dataset from an lmdb file. Adapted from `TAPE <https://github.com/songlab-cal/tape/blob/master/tape/datasets.py>`_.

    :param data_file: path to LMDB file containing dataset
    :type data_file: Union[str, Path]
    """

    def __init__(self, data_file, transform=None):
        """constructor

        """
        if type(data_file) is list:
            if len(data_file) != 1:
                raise RuntimeError("Need exactly one filepath for lmdb")
            data_file = data_file[0]

        self.data_file = Path(data_file).absolute()
        if not self.data_file.exists():
            raise FileNotFoundError(self.data_file)

        env = lmdb.open(str(self.data_file), max_readers=1, readonly=True,
                        lock=False, readahead=False, meminit=False)

        with env.begin(write=False) as txn:
            self._num_examples = int(txn.get(b'num_examples'))
            self._serialization_format = \
                txn.get(b'serialization_format').decode()
            self._id_to_idx = deserialize(
                txn.get(b'id_to_idx'), self._serialization_format)

        self._env = env
        self._transform = transform

    def __len__(self) -> int:
        return self._num_examples

    def get(self, id: str):
        idx = self.id_to_idx(id)
        return self[idx]

    def id_to_idx(self, id: str):
        if id not in self._id_to_idx:
            raise IndexError(id)
        idx = self._id_to_idx[id]
        return idx

    def ids_to_indices(self, ids):
        return [self.id_to_idx(id) for id in ids]

    def ids(self):
        return list(self._id_to_idx.keys())

    def __getitem__(self, index: int):
        if not 0 <= index < self._num_examples:
            raise IndexError(index)

        with self._env.begin(write=False) as txn:

            compressed = txn.get(str(index).encode())
            buf = io.BytesIO(compressed)
            with gzip.GzipFile(fileobj=buf, mode="rb") as f:
                serialized = f.read()
            item = deserialize(serialized, self._serialization_format)

        # Items that start with prefix atoms are assumed to be a dataframe.
        for x in item.keys():
            if x.startswith('atoms'):
                item[x] = pd.DataFrame(**item[x])

        if self._transform:
            item = self._transform(item)
        if 'file_path' not in item:
            item['file_path'] = str(self.data_file)
        if 'id' not in item:
            item['id'] = str(index)
        return item


class PDBDataset(Dataset):
    """
    Creates a dataset from a list of PDB files.

    :param file_list: path to LMDB file containing dataset
    :type file_list: list[Union[str, Path]]
    :param transform: transformation function for data augmentation, defaults to None
    :type transform: function, optional
    """

    def __init__(self, file_list, transform=None):
        """constructor

        """
        self._file_list = [Path(x).absolute() for x in file_list]
        self._num_examples = len(self._file_list)
        self._transform = transform

    def __len__(self) -> int:
        return self._num_examples

    def __getitem__(self, index: int):
        if not 0 <= index < self._num_examples:
            raise IndexError(index)

        file_path = self._file_list[index]

        item = {
            'atoms': fo.bp_to_df(fo.read_any(file_path)),
            'id': file_path.name
        }
        if self._transform:
            item = self._transform(item)
        return item


class SilentDataset(IterableDataset):
    """
    Creates a dataset from rosetta silent files. Can either use a directory of silent files, or a path to one.

    :param file_list: list containing paths to silent files
    :type file_list: list[Union[str, Path]]
    :param transform: transformation function for data augmentation, defaults to None
    :type transform: function, optional
    """

    def __init__(self, file_list, transform=None):
        """constructor

        """

        if not importlib.util.find_spec("rosetta") is not None:
            raise RuntimeError(
                'Need to install pyrosetta to process silent files.')

        with contextlib.redirect_stdout(None):
            self.pyrosetta = importlib.import_module('pyrosetta')
            self.pyrpose = importlib.import_module(
                'pyrosetta.rosetta.core.pose')
            self.pyrps = importlib.import_module(
                'pyrosetta.rosetta.core.import_pose.pose_stream')
            self.pyrosetta.init("-mute all")

        self._file_list = [Path(x).absolute() for x in file_list]
        self._scores = ar.Scores(self._file_list)
        self._transform = transform

        self._num_examples = len(self._scores)

    def __len__(self) -> int:
        return len(self._scores)

    def __iter__(self):
        for silent_file in self._file_list:
            pis = self.pyrps.SilentFilePoseInputStream(str(silent_file))
            while pis.has_another_pose():
                pose = self.pyrosetta.Pose()
                pis.fill_pose(pose)

                item = {
                    'atoms': self._pose_to_df(pose),
                    'id': self.pyrpose.tag_from_pose(pose),
                    'file_path': str(silent_file),
                }
                item['scores'] = self._scores(item)

                if self._transform:
                    item = self._transform(item)

                yield item

    def _pose_to_df(self, pose):
        """
        Convert pyrosetta representation to pandas dataframe representation.
        """
        name = pose.pdb_info().name()
        string_stream = self.pyrosetta.rosetta.std.ostringstream()
        pose.dump_pdb(string_stream)
        f = io.StringIO(string_stream.str())
        parser = Bio.PDB.PDBParser(QUIET=True)
        bp = parser.get_structure(name, f)
        return fo.bp_to_df(bp)


class XYZDataset(Dataset):
    """
    Creates a dataset from list of XYZ files.

    :param file_list: list containing paths to xyz files
    :type file_list: list[Union[str, Path]]
    :param transform: transformation function for data augmentation, defaults to None
    :type transform: function
    :param gdb: whether to add new energies with subtracted thermochemical energies (for SMP dataset), defaults to False
    :type gdb: bool, optional
    """

    def __init__(self, file_list, transform=None, gdb=False):
        """constructor

        """
        self._file_list = [Path(x) for x in file_list]
        self._num_examples = len(self._file_list)
        self._transform = transform
        self._gdb = gdb

    def __len__(self) -> int:
        return self._num_examples

    def __getitem__(self, index: int):
        if not 0 <= index < self._num_examples:
            raise IndexError(index)

        file_path = self._file_list[index]
        bp = fo.read_xyz(file_path, gdb=self._gdb)
        if self._gdb:
            bp, data, freq, smiles, inchi = bp
        df = fo.bp_to_df(bp)

        item = {
            'atoms': df,
            'id': bp.id,
            'file_path': str(file_path),
        }
        if self._gdb:
            item['labels'] = data
            item['freq'] = freq
        if self._transform:
            item = self._transform(item)
        return item


class SDFDataset(Dataset):
    """
    Creates a dataset from directory of SDF files.

    :param file_list: list containing paths to SDF files. Assumes one structure per file.
    :type file_list: list[Union[str, Path]]
    :param transform: transformation function for data augmentation, defaults to None
    :type transform: function, optional
    :param read_bonds: flag for whether to process bond information from SDF, defaults to False
    :type read_bonds: bool, optional
    """

    def __init__(self, file_list, transform=None, read_bonds=False):
        """constructor

        """
        self._file_list = [Path(x) for x in file_list]
        self._num_examples = len(self._file_list)
        self._transform = transform
        self._read_bonds = read_bonds

    def __len__(self) -> int:
        return self._num_examples

    def __getitem__(self, index: int):
        if not 0 <= index < self._num_examples:
            raise IndexError(index)
        # Read biopython structure
        file_path = self._file_list[index]
        structure = fo.read_sdf(str(file_path), sanitize=False,
                                add_hs=False, remove_hs=False)
        # assemble the item (no bonds)
        item = {
            'atoms': fo.bp_to_df(structure),
            'id': structure.id,
            'file_path': str(file_path),
        }
        # Add bonds if included
        if self._read_bonds:
            mol = fo.read_sdf_to_mol(str(file_path), sanitize=False,
                                     add_hs=False, remove_hs=False)
            bonds_df = fo.get_bonds_list_from_mol(mol[0])
            item['bonds'] = bonds_df
        if self._transform:
            item = self._transform(item)
        return item


def serialize(x, serialization_format):
    """
    Serializes dataset `x` in format given by `serialization_format` (pkl, json, msgpack).
    """
    if serialization_format == 'pkl':
        # Pickle
        # Memory efficient but brittle across languages/python versions.
        return pkl.dumps(x)
    elif serialization_format == 'json':
        # JSON
        # Takes more memory, but widely supported.
        serialized = json.dumps(
            x, default=lambda df: json.loads(
                df.to_json(orient='split', double_precision=6))).encode()
    elif serialization_format == 'msgpack':
        # msgpack
        # A bit more memory efficient than json, a bit less supported.
        serialized = msgpack.packb(
            x, default=lambda df: df.to_dict(orient='split'))
    else:
        raise RuntimeError('Invalid serialization format')
    return serialized


def deserialize(x, serialization_format):
    """
    Deserializes dataset `x` assuming format given by `serialization_format` (pkl, json, msgpack).
    """
    if serialization_format == 'pkl':
        return pkl.loads(x)
    elif serialization_format == 'json':
        serialized = json.loads(x)
    elif serialization_format == 'msgpack':
        serialized = msgpack.unpackb(x)
    else:
        raise RuntimeError('Invalid serialization format')
    return serialized


def get_file_list(input_path, filetype):
    if filetype == 'lmdb':
        file_list = [input_path]
    else:
        file_list = fi.find_files(input_path, fo.patterns[filetype])
    return file_list


def load_dataset(file_list, filetype, transform=None, include_bonds=False):
    """
    Load files in file_list into corresponding dataset object. All files should be of type filetype.

    :param file_list: List containing paths to silent files. Assumes one structure per file.
    :type file_list: list[Union[str, Path]]
    :param filetype: Type of dataset. Allowable types are 'lmdb', 'pdb', 'silent', 'sdf', 'xyz', 'xyz-gdb'.
    :type filetype: str
    :param transform: transformation function for data augmentation, defaults to None
    :type transform: function, optional
    :param include_bonds: flag for whether to process bond information for small molecules, defaults to False
    :type include_bonds: bool, optional

    :return: Pytorch Dataset containing data
    :rtype: torch.utils.data.Dataset
    """
    if type(file_list) != list:
        file_list = get_file_list(file_list, filetype)

    if filetype == 'lmdb':
        dataset = LMDBDataset(file_list, transform=transform)
    elif filetype == 'pdb':
        dataset = PDBDataset(file_list, transform=transform)
    elif filetype == 'silent':
        dataset = SilentDataset(file_list, transform=transform)
    elif filetype == 'sdf':
        # TODO: Make read_bonds parameter part of transform.
        dataset = SDFDataset(file_list, transform=transform,
                             read_bonds=include_bonds)
    elif filetype == 'xyz':
        dataset = XYZDataset(file_list, transform=transform)
    elif filetype == 'xyz-gdb':
        # TODO: Make gdb parameter part of transform.
        dataset = XYZDataset(file_list, transform=transform, gdb=True)
    else:
        raise RuntimeError(f'Unrecognized filetype {filetype}.')
    return dataset


def make_lmdb_dataset(dataset, output_lmdb,
                      filter_fn=None, serialization_format='json',
                      include_bonds=False):
    """
    Make an LMDB dataset from an input dataset.

    :param dataset: Input dataset to convert
    :type dataset: torch.utils.data.Dataset
    :param output_lmdb: Path to output LMDB.
    :type output_lmdb: Union[str, Path]
    :param filter_fn: Filter to decided if removing files.
    :type filter_fn: lambda x -> True/False
    :param serialization_format: How to serialize an entry.
    :type serialization_format: 'json', 'msgpack', 'pkl'
    :param include_bonds: Include bond information (only available for SDF yet).
    :type include_bonds: bool
    """

    num_examples = len(dataset)

    logger.info(f'{num_examples} examples')

    env = lmdb.open(str(output_lmdb), map_size=int(1e11))

    with env.begin(write=True) as txn:

        id_to_idx = {}
        i = 0
        for x in tqdm.tqdm(dataset, total=num_examples):
            if filter_fn is not None and filter_fn(x):
                continue
            buf = io.BytesIO()
            with gzip.GzipFile(fileobj=buf, mode="wb", compresslevel=6) as f:
                f.write(serialize(x, serialization_format))
            compressed = buf.getvalue()
            result = txn.put(str(i).encode(), compressed, overwrite=False)
            if not result:
                raise RuntimeError(f'LMDB entry {i} in {str(output_lmdb)} '
                                   'already exists')

            id_to_idx[x['id']] = i
            i += 1

        txn.put(b'num_examples', str(i).encode())
        txn.put(b'serialization_format', serialization_format.encode())
        txn.put(b'id_to_idx', serialize(id_to_idx, serialization_format))

        
def extract_coordinates_as_numpy_arrays(dataset, indices=None, atom_frames=['atoms'], drop_elements=[]):
    """Convert the molecules from a dataset to a dictionary of numpy arrays.
       Labels are not processed; they are handled differently for every dataset.
       
    :param dataset: LMDB dataset from which to extract coordinates.
    :type dataset: torch.utils.data.Dataset
    :param indices: Indices of the items for which to extract coordinates.
    :type indices: numpy.array
    :param atom_frames: keys for the frames that contain the atoms to be written.
    :type atom_frames: [str]

    :return: Dictionary of numpy arrays with number of atoms, charges, and positions
    :rtype: dict       
    """
    # Size of the dataset
    if indices is None:
        indices = np.arange(len(dataset))
    else:
        assert len(dataset) > max(indices)
    num_items = len(indices)

    # Calculate number of atoms for each molecule
    num_atoms = []
    for idx in indices:
        item = dataset[idx]
        atoms = pd.concat([item[frame] for frame in atom_frames])
        keep = np.array([el not in drop_elements for el in atoms['element']])
        num_atoms.append(sum(keep))

    # All charges and position arrays have the same size
    arr_size  = np.max(num_atoms)
    charges   = np.zeros([num_items,arr_size])
    positions = np.zeros([num_items,arr_size,3])
    # For each molecule and each atom...
    for j,idx in enumerate(indices):
        item = dataset[idx]
        # concatenate atoms from all desired frames
        all_atoms = [item[frame] for frame in atom_frames] 
        atoms = pd.concat(all_atoms, ignore_index=True)
        # only keep atoms that are not one of the elements to drop
        keep = np.array([el not in drop_elements for el in atoms['element']])
        atoms_to_keep = atoms[keep].reset_index(drop=True)
        # write per-atom data to arrays
        for ia in range(num_atoms[j]):
            element = atoms_to_keep['element'][ia].title()
            charges[j,ia] = fo.atomic_number[element]
            positions[j,ia,0] = atoms_to_keep['x'][ia] 
            positions[j,ia,1] = atoms_to_keep['y'][ia]
            positions[j,ia,2] = atoms_to_keep['z'][ia]
            
    # Create a dictionary with all the arrays
    numpy_dict = {'index':indices, 'num_atoms':num_atoms,
                  'charges':charges, 'positions':positions}
    
    return numpy_dict

def combine_datasets(dataset_list, output_lmdb, filter_fn=None, serialization_format='json'):
    """
    Combine list of datasets (in any format) to single LMDB dataset.

    :param dataset_list: List of input datasets
    :type dataset_list: List[torch.utils.data.Dataset]
    :param output_lmdb: Path to output LMDB.
    :type output_lmdb: Union[str, Path]
    :param filter_fn: Filter to decided if removing files.
    :type filter_fn: lambda x -> True/False
    :param serialization_format: How to serialize an entry.
    :type serialization_format: 'json', 'msgpack', 'pkl'
    """
    
    num_examples = np.sum([len(d) for d in dataset_list])

    logger.info(f'{num_examples} examples in combined dataset')

    env = lmdb.open(str(output_lmdb), map_size=int(1e11))

    with env.begin(write=True) as txn:

        id_to_idx = {}
        i = 0
        
        for dset in dataset_list:
            for x in tqdm.tqdm(dset, initial=i, total=num_examples):
                if filter_fn is not None and filter_fn(x):
                    continue
                buf = io.BytesIO()
                with gzip.GzipFile(fileobj=buf, mode="wb", compresslevel=6) as f:
                    f.write(serialize(x, serialization_format))
                compressed = buf.getvalue()
                result = txn.put(str(i).encode(), compressed, overwrite=False)
                if not result:
                    raise RuntimeError(f'LMDB entry {i} in {str(output_lmdb)} '
                                    'already exists')

                id_to_idx[x['id']] = i
                i += 1

        txn.put(b'num_examples', str(i).encode())
        txn.put(b'serialization_format', serialization_format.encode())
        txn.put(b'id_to_idx', serialize(id_to_idx, serialization_format))

def download_dataset(name, out_path):
    """Download an ATOM3D dataset in LMDB format. Available datasets are SMP, PIP, RES, MSP, LBA, LEP, PSR, RSR. Please see `FAQ <datasets target>`_ or `atom3d.ai <atom3d.ai>`_ for more details on each dataset.

    :param name: Three-letter code for dataset (not case-sensitive).
    :type name: str
    :param out_path: Path to directory in which to save downloaded dataset.
    :type out_path: str
    """

    def _hook(t):
        """from https://github.com/tqdm/tqdm/blob/master/examples/tqdm_wget.py
        """
        last_b = [0]

        def update_to(b=1, bsize=1, tsize=None):
            """
            b  : int, optional
                Number of blocks transferred so far [default: 1].
            bsize  : int, optional
                Size of each block (in tqdm units) [default: 1].
            tsize  : int, optional
                Total size (in tqdm units). If [default: None] or -1,
                remains unchanged.
            """
            if tsize not in (None, -1):
                t.total = tsize
            displayed = t.update((b - last_b[0]) * bsize)
            last_b[0] = b
            return displayed

        return update_to

    name = name.lower()
    if name == 'smp':
        link = '13MT_f86so0fm6TOtzhW2Qy9ubVQo6UiU'
    elif name == 'pip':
        link = '1D4gMdJEz-6hzSc7_QQ2CF1K-anR4mO8T'
    elif name == 'res':
        link = '1XgZ19YYwloHxEtZUk78PLVzHipFkqIm5'
    elif name == 'msp':
        link = '15rojYF-UjNnqoD8BnNpFtoxVZu64Y7FL'
    elif name == 'lba':
        link = '1CGCRj3IwbT0HNSHIqQ46-o2n1CmGOnwK'
    elif name == 'lep':
        link = '15A85q2h6C1WFKjVttv6sInFNnB5z7Ha7'
    elif name == 'psr':
        link = '1rvxf9JKTq0OvU3QLkxNYomfyXg5sd2CO'
    elif name == 'rsr':
        link = '1rlQ8BmyamMud2TZkcFGy_raz9iI1-KMm'
    else:
        print('Invalid dataset name specified. Possible values are {SMP, PIP, RES, MSP, LBA, LEP, PSR, RSR}')

    # f_out = os.path.join(out_path, name + '.lmdb')
    # with tqdm.tqdm(unit='B', unit_scale=True, unit_divisor=1024, miniters=1, desc=f_out) as t:  # all optional kwargs
    #     urllib.request.urlretrieve(link, filename=f_out,
    #                     reporthook=_hook(t), data=None)

    cmd = f"wget --load-cookies /tmp/cookies.txt \"https://docs.google.com/uc?export=download&confirm=$(wget --quiet --save-cookies /tmp/cookies.txt --keep-session-cookies --no-check-certificate 'https://docs.google.com/uc?export=download&id={link}' -O- | sed -En 's/.*confirm=([0-9A-Za-z_]+).*/\\1\\n/p'  | tr -d \"n\")&id={link}\" -O {name}.tar.gz"
    subprocess.call(cmd, shell=True)
    cmd2 = f"tar xzvf {name}.tar.gz"
    subprocess.call(cmd2, shell=True)




"""Functions for splitting data into test, validation, and training sets."""
from functools import partial

import numpy as np
import torch

import atom3d.util.log as log

logger = log.get_logger('splits')


def split(dataset, indices_train, indices_val, indices_test):
    """Split a dataset into train, validation, and test datasets according to specified indices.

    :param dataset: Dataset to split.
    :type dataset: Dataset
    :param indices_train: List of indices comprising training set.
    :type indices_train: List[int]
    :param indices_val: List of indices comprising validation set.
    :type indices_val: List[int]
    :param indices_test: List of indices comprising test set.
    :type indices_test: List[int]
    :return: Tuple of train, validation, and test datasets
    :rtype: Tuple[Dataset]
    """    
    train_dataset = torch.utils.data.Subset(dataset, indices_train)
    val_dataset = torch.utils.data.Subset(dataset, indices_val)
    test_dataset = torch.utils.data.Subset(dataset, indices_test)

    logger.info(f'Size of the training set: {len(indices_train):}')
    logger.info(f'Size of the validation set: {len(indices_val):}')
    logger.info(f'Size of the test set: {len(indices_test):}')

    return train_dataset, val_dataset, test_dataset


def read_split_file(split_file):
    """
    Read text file with pre-defined split, returning list of examples.

    One example per row in text file.
    """
    with open(split_file) as f:
        # file may contain integer indices or string identifiers (e.g. PDB
        # codes)
        lines = f.readlines()
        try:
            split = [int(x.strip()) for x in lines]
        except ValueError:
            split = [x.strip() for x in lines]
    return split


####################################
# split randomly
####################################

def split_randomly(dataset, train_split=None, val_split=0.1, test_split=0.1, random_seed=0):
    """Split a dataset into train, validation and test datasets at random.

    :param dataset: Dataset to split.
    :type dataset: Dataset
    :param train_split: Proportion of data used for training. If None, use all data not in validation or test, defaults to None.
    :type train_split: float, optional
    :param val_split: Proportion of data used for validation, defaults to 0.1
    :type val_split: float, optional
    :param test_split: Proportion of data used for testing, defaults to 0.1
    :type test_split: float, optional
    :param random_seed: Random seed for splitting, defaults to 0
    :type random_seed: int, optional
    :return: Tuple of train, validation, and test datasets
    :rtype: Tuple[Dataset]
    """    

    # Initialize the indices
    num_indices = len(dataset)
    indices = np.arange(num_indices, dtype=int)
    logger.info(f'Splitting dataset with {num_indices:} entries.')

    # Calculate the numbers of elements per split
    num_val = int(np.floor(val_split * num_indices))
    num_test = int(np.floor(test_split * num_indices))
    if train_split is not None:
        num_train = int(np.floor(train_split * num_indices))
    else:
        num_train = num_indices - num_val - num_test

    # Shuffle the dataset indices
    np.random.seed(random_seed)
    np.random.shuffle(indices)

    # Determine the indices of each split
    indices_train = indices[:num_train]
    indices_val = indices[num_train:(num_train + num_val)]
    indices_test = indices[(num_train + num_val):(num_train + num_val + num_test)]

    return split(dataset, indices_train, indices_val, indices_test)


####################################
# split by group
####################################

def split_by_group(dataset, value_fn, train_values, val_values, test_values):
    """Splits data into train, validation, and test dataset using a value function that maps each data element to a value (or group identifier). These are then used to assign elements to the appropriate splits based on pre-defined lists of values to include in each split.

    :param dataset: Dataset to split.
    :type dataset: Dataset
    :param value_fn: Arbitrary function mapping each data element to a value or group identifier.
    :type value_fn: function
    :param train_values: List of values to include in training set.
    :type train_values: List
    :param val_values: List of values to include in validation set.
    :type val_values: List
    :param test_values: List of values to include in test set.
    :type test_values: List
    :return: Tuple of train, validation, and test datasets
    :rtype: Tuple[Dataset]
    """    

    values = [value_fn(x) for x in dataset]

    # Determine the indices of each split
    indices_train = [i for i,x in enumerate(values) if x in train_values]
    indices_val = [i for i,x in enumerate(values) if x in val_values]
    indices_test = [i for i,x in enumerate(values) if x in test_values]
    return split(dataset, indices_train, indices_val, indices_test)


####################################
# split by group size
####################################


def split_by_group_size(dataset, value_fn, val_split=0.1, test_split=0.1):
    """Splits data into train, validation, and test dataset using a value function that maps each data element to a value (or group identifier). Elements are grouped by the value returned by this value function, and groups are sorted by size. 
    The groups are then added first to train, then to validation, then to test splits in order of group size, so that the largest groups (i.e. most common examples)  are in train and the smallest groups (i.e. less common examples) are in test. 
    Each split is filled iteratively up to the sizes specified by ``val_split`` and ``test_split``.

    :param dataset: Dataset to split.
    :type dataset: Dataset
    :param value_fn: Arbitrary function mapping each data element to a value or group identifier.
    :type value_fn: function
    :param val_split: Proportion of data used for validation, defaults to 0.1
    :type val_split: float, optional
    :param test_split: Proportion of data used for testing, defaults to 0.1
    :type test_split: float, optional
    :return: Tuple of train, validation, and test datasets.
    :rtype: Tuple[Dataset]
    """    

    values = [value_fn(x) for x in dataset]

    logger.info(f'Splitting dataset with {len(values):} entries.')

    # Calculate the target sizes of the splits
    dataset_size = len(values)
    all_indices = np.arange(dataset_size)
    test_size = test_split * dataset_size
    val_size = val_split * dataset_size
    train_size = dataset_size - val_size - test_size

    # Order the scaffolds from common to uncommon 
    unique_values, counts = np.unique(values, return_counts=True)
    order = np.argsort(counts)[::-1]
    values_ordered = unique_values[order]

    # Initialize index lists
    indices_train = []
    indices_val = []
    indices_test = []
    # Initialize counters for scaffolds in each set
    num_sc_train = 0
    num_sc_val = 0
    num_sc_test = 0

    # Go through the scaffolds from common to uncommon 
    # and fill the training, validation, and test sets
    for sc in values_ordered:
        # Get all indices of the current scaffold
        curr_group = all_indices[np.array(values) == sc].tolist()
        # ... and add them to their dataset
        if len(indices_train) < train_size:
            indices_train += curr_group
            num_sc_train += 1
        elif len(indices_val) < val_size:
            indices_val += curr_group
            num_sc_val += 1
        else:
            indices_test += curr_group
            num_sc_test += 1

    # Report number of scaffolds in each set
    logger.info(f'Groups in the training set: {int(num_sc_train):}')
    logger.info(f'Groups in the validation set: {int(num_sc_val):}')
    logger.info(f'Groups in the test set: {int(num_sc_test):}')

    # Report number of scaffolds in each set
    return split(dataset, indices_train, indices_val, indices_test)

    
####################################
# frequently used specific splits
####################################

split_by_year = partial(split_by_group, value_fn=lambda x: x['year'])

split_by_scaffold = partial(split_by_group_size, value_fn=lambda x: x['scaffold'])



"""
Common filtering functions.

These all are applied to individual atom dataframes, and remove entries from that dataframe as necessary.
"""
import Bio.PDB.Polypeptide as Poly
import pandas as pd
import scipy.spatial as ss
import numpy as np


def standard_residue_filter(df):
    """
    Filter out non-standard residues.

    :param df: dataframe to filter against.
    :type df: atoms dataframe.

    :return: same dataframe, but with only with atoms corresponding to standard residues left.
    :rtype: atoms dataframe.
    """
    residues = df[['structure', 'model', 'chain', 'residue', 'resname']] \
        .drop_duplicates()
    sel = residues['resname'].apply(
        lambda x: Poly.is_aa(x, standard=True))

    residues['to_keep'] = sel
    residues_to_keep = residues.set_index(
        ['structure', 'model', 'chain', 'residue', 'resname'])['to_keep']
    to_keep = residues_to_keep.loc[df.set_index(
        ['structure', 'model', 'chain', 'residue', 'resname']).index]
    return df[to_keep.values]


def first_model_filter(df):
    """
    Remove anything beyond first model in structure.

    :param df: dataframe to filter against.
    :type df: atoms dataframe.

    :return: same dataframe, but with only with atoms corresponding to first model left.
    :rtype: atoms dataframe.
    """
    models = df[['structure', 'model']].drop_duplicates()
    models = models.sort_values(['structure', 'model'])

    models['to_keep'] = ~models['structure'].duplicated()
    models_to_keep = models.set_index(['structure', 'model'])

    to_keep = models_to_keep.loc[df.set_index(['structure', 'model']).index]
    return df[to_keep.values]


def first_chain_filter(df):
    """
    Remove anything beyond first model/chain in structure.

    :param df: dataframe to filter against.
    :type df: atoms dataframe.

    :return: same dataframe, but with only with atoms corresponding to the first chain of the first model left.
    :rtype: atoms dataframe.
    """

    chains = df[['structure', 'model', 'chain']].drop_duplicates()
    chains = chains.sort_values(['structure', 'model', 'chain'])

    chains['to_keep'] = ~chains['structure'].duplicated()
    chains_to_keep = chains.set_index(['structure', 'model', 'chain'])

    to_keep = \
        chains_to_keep.loc[df.set_index(['structure', 'model', 'chain']).index]
    return df[to_keep.values]

def single_chain_filter(df):
    """
    Remove anything beyond first model/chain in structure.

    :param df: dataframe to filter against.
    :type df: atoms dataframe.

    :return: same dataframe if structure has only one chain, otherwise empty dataframe.
    :rtype: atoms dataframe.
    """

    chains = df[['structure', 'model', 'chain']].drop_duplicates()
    chains = chains.sort_values(['structure', 'model', 'chain'])

    chains['to_keep'] = ~chains['structure'].duplicated(keep=False)
    chains_to_keep = chains.set_index(['structure', 'model', 'chain'])

    to_keep = \
        chains_to_keep.loc[df.set_index(['structure', 'model', 'chain']).index]
    return df[to_keep.values]

def identity_filter(df):
    """
    Leave atoms dataframe unchanged.

    :param df: dataframe to filter against.
    :type df: atoms dataframe.

    :return: same dataframe.
    :rtype: atoms dataframe.
    """
    return df

def distance_filter(df, pos, dist):
    """Returns dataframe containing all atoms within ``dist`` of query positions in ``pos``.

    :param df: Input structure dataframe
    :type df: pandas.DataFrame
    :param pos: x-y-z positions of query points. For N query points, should be array of shape N x 3
    :type pos: array-like
    :param dist: Distance in Angstrom to search for neighbors.
    :type dist: float
    :return: New dataframe containing only atoms within ``dist`` of query position
    :rtype: pandas.DataFrame
    """    
    if pos.ndim == 1:
        pos = pos[np.newaxis, :]
    kd_tree = ss.KDTree(df[['x','y','z']].values)
    nn_pt_idx = kd_tree.query_ball_point(pos, r=dist, p=2.0)
    nn_pt_idx = list(set([k for l in nn_pt_idx for k in l]))
    nn_df = df.iloc[nn_pt_idx].reset_index(drop=True)
    return nn_df

def compose(f, g):
    """
    Compose two filter f and g.

    :param f: Outer filter function.
    :type f: filter function.
    :param g: Inner filter function.
    :type g: filter function.

    :return: lambda x: f(g(x))
    :rtype: filter function.
    """

    def filter_fn(df):
        df = g(df)
        if len(df) > 0:
            return f(df)
        else:
            return df
    return filter_fn


def form_filter_against_list(against, column):
    """
    Create filter against a list, keeping only items with values in that list.

    :param against: values that we will keep when they are found.
    :type against: list.
    :param column: dataframe column that we will extract values from.
    :type column: str.

    :return: function that implements the specified filter.
    :rtype: filter function.
    """

    def filter_fn(df):
        to_keep = {}
        for e, ensemble in df.groupby([column]):
            to_keep[e] = e in against
        to_keep = pd.Series(to_keep)[df[column]]
        return df[to_keep.values]

    return filter_fn

def filter_to_transform(filter_fn, df_name='atoms'):
    """Create transform function (which operates on dataset items) from filter function (which operates on dataframes). By default, applies filter_fn to ``atoms`` dataframe, but a different dataframe can be specified optionally.

    :param filter_fn: Arbitrary filter function that takes in a dataframe and returns filtered dataframe
    :type filter_fn: filter function
    :param df_name: Name of dataframe to apply filter_fn to, defaults to 'atoms'
    :type df_name: str, optional
    """    
    def transform_fn(item):
        item[df_name] = filter_fn(item['atoms'])
        return item
    return transform_fn


"""Other operations for sharded datasets."""
import logging
import os
import random

import tqdm

import atom3d.shard.shard as sh
import atom3d.util.formats as dt


def filter_sharded(input_sharded, output_sharded, filter_fn, shuffle_buffer=0):
    """Filter sharded dataset to new sharded dataset, using provided filter."""
    logging.basicConfig(format='%(asctime)s %(levelname)s %(process)d: ' +
                        '%(message)s',
                        level=logging.INFO)

    if not os.path.exists(os.path.dirname(output_sharded.path)):
        os.makedirs(os.path.dirname(output_sharded.path))

    input_num_shards = input_sharded.get_num_shards()

    # We will just map to tmp, then reshard.
    tmp_path = output_sharded.get_prefix() + f'_tmp@{input_num_shards:}'
    tmp_sharded = sh.Sharded(tmp_path, input_sharded.get_keys())

    logging.info(f'Filtering {input_sharded.path:} to {output_sharded.path:}')
    # Apply filter.
    for shard_num in tqdm.trange(input_num_shards):
        df = input_sharded.read_shard(shard_num)
        if len(df) > 0:
            df = filter_fn(df)
        tmp_sharded._write_shard(shard_num, df)

    num_input_structures = input_sharded.get_num_keyed()
    num_output_structures = tmp_sharded.get_num_keyed()
    logging.info(f'After filtering, have {num_output_structures:} / '
                 f'{num_input_structures:} left.')
    reshard(tmp_sharded, output_sharded, shuffle_buffer)
    tmp_sharded.delete_files()


def reshard(input_sharded, output_sharded, shuffle_buffer=0):
    """
    Rebalance dataset, optionally shuffling.

    If shuffle_buffer is not 0, then we perform a streaming shuffle across
    shuffle_buffer number of output shards.
    """
    dirname = os.path.dirname(output_sharded.path)
    if not os.path.exists(dirname) and dirname != '':
        os.makedirs(dirname, exist_ok=True)

    num_structures = input_sharded.get_num_keyed()
    output_num_shards = output_sharded.get_num_shards()
    input_num_shards = input_sharded.get_num_shards()

    shard_ranges = sh._get_shard_ranges(num_structures, output_num_shards)
    shard_sizes = shard_ranges[:, 1] - shard_ranges[:, 0]

    if shuffle_buffer != 0:
        buffer_size = shuffle_buffer * shard_sizes[0]
    else:
        buffer_size = 1

    t = tqdm.trange(output_num_shards)
    next_output_shard_num, next_input_shard_num = 0, 0
    to_write, to_consume = [], []
    df = None
    while True:
        while len(to_consume) < buffer_size and \
                (next_input_shard_num != input_num_shards):
            # Read next shard if need more examples.
            df = input_sharded.read_shard(next_input_shard_num)
            to_consume += [y for (_, y) in
                           dt.split_df(df, input_sharded.get_keys())]

            if shuffle_buffer is not None:
                random.shuffle(to_consume)
            next_input_shard_num += 1

        if len(to_consume) != 0:
            to_write.append(to_consume.pop(0))

        if len(to_write) == shard_sizes[next_output_shard_num]:
            # Write output shard if have number needed.

            if len(to_write) == 0:
                # Insert empty dataframe if nothing to write.
                to_write = [df.iloc[0:0]]

            output_sharded._write_shard(next_output_shard_num,
                                        dt.merge_dfs(to_write))
            to_write = []
            next_output_shard_num += 1
            t.update(1)

            if next_output_shard_num == output_num_shards:
                break


def rekey(input_sharded, output_sharded, shuffle_buffer=0):
    """Rekey dataset."""
    dirname = os.path.dirname(output_sharded.path)
    if not os.path.exists(dirname) and dirname != '':
        os.makedirs(dirname, exist_ok=True)

    input_num_shards = input_sharded.get_num_shards()

    # We will just map to tmp, then reshard.
    tmp_path = output_sharded.get_prefix() + f'_tmp@{input_num_shards:}'
    tmp_sharded = sh.Sharded(tmp_path, output_sharded.get_keys())

    for shard_num in tqdm.trange(input_num_shards):
        df = input_sharded.read_shard(shard_num)
        tmp_sharded._write_shard(shard_num, df)

    num_input_structures = input_sharded.get_num_keyed()
    num_output_structures = tmp_sharded.get_num_keyed()
    logging.info(f'After rekey-ing, have {num_output_structures:} keyed, '
                 f'from {num_input_structures:} originally.')
    reshard(tmp_sharded, output_sharded, shuffle_buffer)
    tmp_sharded.delete_files()



import click
import logging
import sys

import atom3d.datasets.datasets as da
import atom3d.util.file as fi
import atom3d.util.formats as fo

logger = logging.getLogger(__name__)


@click.command()
@click.argument('input_dir', type=click.Path(exists=True))
@click.argument('output_lmdb', type=click.Path(exists=False))
@click.option('-f', '--filetype', type=click.Choice(['pdb', 'silent', 'xyz', 'xyz-gdb']),
              default='pdb')
@click.option('-sf', '--serialization_format',
              type=click.Choice(['msgpack', 'pkl', 'json']),
              default='json')
@click.option('--score_path', type=click.Path(exists=True))
def main(input_dir, output_lmdb, filetype, score_path, serialization_format):
    """Script wrapper to make_lmdb_dataset to create LMDB dataset."""
    logging.basicConfig(stream=sys.stdout,
                        format='%(asctime)s %(levelname)s %(process)d: ' +
                        '%(message)s',
                        level=logging.INFO)

    logger.info(f'filetype: {filetype}')
    if filetype == 'xyz-gdb':
        fileext = 'xyz'
    else:
        fileext = filetype
    file_list = fi.find_files(input_dir, fo.patterns[fileext])
    file_list.sort()
    logger.info(f'Found {len(file_list)} files.')

    dataset = da.load_dataset(file_list, filetype)
    da.make_lmdb_dataset(
        dataset, output_lmdb, serialization_format=serialization_format)


if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd

import atom3d.util.vectors as vec

import itertools
import operator

class dotdict(dict):
    """ dot.notation access to dictionary attributes """
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def get_center(pos):
    """ Return the center coordinate. """
    return pos.mean().to_numpy().astype(np.float32)


def get_max_distance_from_center(pos, center):
    """ Return maximum distance from center. """
    max_dist = np.sqrt(pos.sub(center).pow(2).sum(1).max())
    return max_dist


def distribute(sequence):
    """
    Enumerate the sequence evenly over the interval (0, 1).

    >>> list(distribute('abc'))
    [(0.25, 'a'), (0.5, 'b'), (0.75, 'c')]
    """
    m = len(sequence) + 1
    for i, x in enumerate(sequence, 1):
        yield i/m, x


def intersperse(*sequences):
    """
    Evenly intersperse the sequences.

    Based on https://stackoverflow.com/a/19293603/4518341

    >>> list(intersperse(range(10), 'abc'))
    [0, 1, 'a', 2, 3, 4, 'b', 5, 6, 7, 'c', 8, 9]
    >>> list(intersperse('XY', range(10), 'abc'))
    [0, 1, 'a', 2, 'X', 3, 4, 'b', 5, 6, 'Y', 7, 'c', 8, 9]
    >>> ''.join(intersperse('hlwl', 'eood', 'l r!'))
    'hello world!'
    """
    distributions = map(distribute, sequences)
    get0 = operator.itemgetter(0)
    for _, x in sorted(itertools.chain(*distributions), key=get0):
        yield x



def rot_mats(config):
    # Generate the rotation matrices.
    if '_rot_mats' not in rot_mats.__dict__:
        _uvs = vec.fibonacci(config.num_directions)
        _ups = vec.generate_all_up_vectors(_uvs, config.num_rolls)
        rot_mats._rot_mats = vec.get_all_rot_mats(_uvs, _ups).astype(np.float32)
    return rot_mats._rot_mats


def gen_rot_matrix(config, random_seed=None, reset_seed=False):
    if reset_seed:
        np.random.seed(random_seed)

    size = grid_size(config)
    true_radius = size * config.resolution / 2.0
    direction = np.random.randint(low=0, high=config.num_directions, dtype=np.int32)
    roll = np.random.randint(low=0, high=config.num_rolls, dtype=np.int32)
    rot_mat = rot_mats(config)[direction][roll]
    return rot_mat


def get_grid(df, center, config, rot_mat=np.eye(3, 3)):
    """
    Generate the 3d grid from coordinate format.
    Args:
        df (pd.DataFrame):
            region to generate grid for.
        center (3x3 np.array):
            center of the grid.
        rot_mat (3x3 np.array):
            rotation matrix to apply to region before putting in grid.
    Returns:
        4-d numpy array representing an occupancy grid where last dimension
        is atom channel.  First 3 dimension are of size radius_ang * 2 + 1.
    """
    size = grid_size(config)
    true_radius = size * config.resolution / 2.0

    # Select valid atoms.
    at = df[['x', 'y', 'z']].values.astype(np.float32)
    elements = df['element'].values

    # Center atoms.
    at = at - center

    # Apply rotation matrix.
    at = np.dot(at, rot_mat)
    at = (np.around((at + true_radius) / config.resolution - 0.5)).astype(np.int16)

    # Prune out atoms outside of grid as well as non-existent atoms.
    sel = np.all(at >= 0, axis=1) & np.all(at < size, axis=1) & (elements != '')
    at = at[sel]

    # Form final grid.
    labels = elements[sel]
    lsel = np.nonzero([_recognized(x, config.element_mapping) for x in labels])
    labels = labels[lsel]
    labels = np.array([config.element_mapping[x] for x in labels], dtype=np.int8)

    grid = np.zeros(grid_shape(config), dtype=np.float32)
    grid[at[lsel, 0], at[lsel, 1], at[lsel, 2], labels] = 1

    return grid


def get_voxels(df, center, config, rot_mat=np.eye(3, 3)):
    """
    Generate the 3d grid from coordinate format.
    Args:
        df (pd.DataFrame):
            region to generate grid for.
        center (3x3 np.array):
            center of the grid.
        rot_mat (3x3 np.array):
            rotation matrix to apply to region before putting in grid.
    Returns:
        4-d numpy array representing an occupancy grid where last dimension
        is atom channel.  First 3 dimension are of size radius_ang * 2 + 1.
    """
    size = grid_size(config)
    true_radius = size * config.resolution / 2.0

    # Select valid atoms.
    at = df[['x', 'y', 'z']].values.astype(np.float32)
    elements = df['element'].values

    # Center atoms.
    at = at - center

    # Apply rotation matrix.
    at = np.dot(at, rot_mat)
    # at = (np.around((at + true_radius) / config.resolution - 0.5)).astype(np.int16)
    bins = np.linspace(-true_radius, true_radius, size+1)
    at_bin_idx = np.digitize(at, bins)

    # Prune out atoms outside of grid as well as non-existent atoms.
    sel = np.all(at_bin_idx > 0, axis=1) & np.all(at_bin_idx < size+1, axis=1) & (elements != '')
    at = at_bin_idx[sel] - 1

    # Form final grid.
    labels = elements[sel]
    lsel = np.nonzero([_recognized(x, config.element_mapping) for x in labels])
    labels = labels[lsel]
    labels = np.array([config.element_mapping[x] for x in labels], dtype=np.int8)

    grid = np.zeros(grid_shape(config), dtype=np.float32)
    grid[at[lsel, 0], at[lsel, 1], at[lsel, 2], labels] = 1

    return grid


def _recognized(x, dict):
    """ If atom type is recognized, return it.  Else, return empty string. """
    if x in dict.keys():
        return x
    else:
        return ''


def num_channels(config):
    return num_element_types(config)


def num_element_types(config):
    return len(config.element_mapping)


def grid_size(config):
    """ Get size of grid. """
    return int(round((config.radius * 2 + 1) / config.resolution))


def grid_shape(config):
    """ Return shape of grid. """
    size = grid_size(config)
    return (size, size, size, num_channels(config))



"""Ensembling operations for shards."""
import pandas as pd

import atom3d.datasets.lep.ensemble as lepe
import atom3d.datasets.msp.ensemble as mspe
import atom3d.datasets.ppi.db5 as db5
import atom3d.datasets.psr.casp as casp
import atom3d.datasets.rsr.ensemble as rsre
import atom3d.datasets.smp.qm9 as qm9
import atom3d.util.formats as dt


def identity_ensembler(pdb_files):
    return {x: None for x in pdb_files}


# An ensembler maps a list of files to an ensemble map (a 2-level dictionary).
#
# An ensemble map takes the form of a 2-level dictionary:
# (e -> (s -> sel))
#
# Where 'e' is the string name of an ensemble, and 's' is the string name of a
# subunit of the ensemble.  'sel' can be one of 1) a filename or 2) a pandas
# DataFrame, and corresponds to the subunit specified by (e, s).
ensemblers = {
    'db5': db5.db5_ensembler,
    'casp': casp.casp_ensembler,
    'lep': lepe.lep_ensembler,
    'rsr': rsre.rsr_ensembler,
    'msp': mspe.msp_ensembler,
    'qm9': qm9.qm9_ensembler,
    'none': identity_ensembler,
}


def parse_ensemble(name, ensemble):
    if ensemble is None:
        df = dt.bp_to_df(dt.read_any(name))
    else:
        df = []
        for subunit, f in ensemble.items():
            if isinstance(f, pd.DataFrame):
                curr = f
            else:
                curr = dt.bp_to_df(dt.read_any(f))

            curr['subunit'] = subunit
            df.append(curr)
        df = pd.concat(df)
        df['ensemble'] = name
    return df



"""File-related utilities."""
import os
from pathlib import Path
import subprocess


def find_files(path, suffix, relative=None):
    """
    Find all files in path with given suffix. =

    :param path: Directory in which to find files.
    :type path: Union[str, Path]
    :param suffix: Suffix determining file type to search for.
    :type suffix: str
    :param relative: Flag to indicate whether to return absolute or relative path.

    :return: list of paths to all files with suffix.
    :rtype: list[Path]
    """
    if not relative:
        find_cmd = r"find {:} -regex '.*\.{:}'".format(path, suffix)
    else:
        find_cmd = r"cd {:}; find . -regex '.*\.{:}' | cut -d '/' -f 2-" \
            .format(path, suffix)
    out = subprocess.Popen(
        find_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        cwd=os.getcwd(), shell=True)
    (stdout, stderr) = out.communicate()
    name_list = stdout.decode().split()
    name_list.sort()
    return [Path(x) for x in name_list]


def get_pdb_code(path):
    """
    Extract 4-character PDB ID code from full path.
    
    :param path: Path to PDB file.
    :type path: str

    :return: PDB filename.
    :rtype: str
    """
    return path.split('/')[-1][:4].lower()


def get_pdb_name(path):
    """
    Extract filename for PDB file from full path.

    :param path: Path to PDB file.
    :type path: str

    :return: PDB filename.
    :rtype: str
    """
    return path.split('/')[-1]



import numpy as np
import scipy.spatial as ss
import torch

import atom3d.util.formats as fo

# PDB atom names -- these include co-crystallized metals
prot_atoms = ['C', 'H', 'O', 'N', 'S', 'P', 'ZN', 'NA', 'FE', 'CA', 'MN', 'NI', 'CO', 'MG', 'CU', 'CL', 'SE', 'F']
# RDKit molecule atom names
mol_atoms = ['C', 'N', 'O', 'S', 'F', 'Si', 'P', 'Cl', 'Br', 'Mg', 'Na',
             'Ca', 'Fe', 'As', 'Al', 'I', 'B', 'V', 'K', 'Tl', 'Yb',
             'Sb', 'Sn', 'Ag', 'Pd', 'Co', 'Se', 'Ti', 'Zn', 'H',  # H?
             'Li', 'Ge', 'Cu', 'Au', 'Ni', 'Cd', 'In', 'Mn', 'Zr',
             'Cr', 'Pt', 'Hg', 'Pb']
# Residue names
residues = ['ALA', 'CYS', 'ASP', 'GLU', 'PHE', 'GLY', 'HIS', 'ILE', 'LYS', 'LEU', 'MET', 'ASN', 'PRO', 'GLN', 'ARG',
                'SER', 'THR', 'VAL', 'TRP', 'TYR']


def prot_df_to_graph(df, feat_col='element', allowable_feats=prot_atoms, edge_dist_cutoff=4.5):
    r"""
    Converts protein in dataframe representation to a graph compatible with Pytorch-Geometric, where each node is an atom.

    :param df: Protein structure in dataframe format.
    :type df: pandas.DataFrame
    :param node_col: Column of dataframe to find node feature values. For example, for atoms use ``feat_col="element"`` and for residues use ``feat_col="resname"``
    :type node_col: str, optional
    :param allowable_feats: List containing all possible values of node type, to be converted into 1-hot node features. 
        Any elements in ``feat_col`` that are not found in ``allowable_feats`` will be added to an appended "unknown" bin (see :func:`atom3d.util.graph.one_of_k_encoding_unk`).
    :type allowable_feats: list, optional
    :param edge_dist_cutoff: Maximum distance cutoff (in Angstroms) to define an edge between two atoms, defaults to 4.5.
    :type edge_dist_cutoff: float, optional

    :return: tuple containing

        - node_feats (torch.FloatTensor): Features for each node, one-hot encoded by values in ``allowable_feats``.

        - edges (torch.LongTensor): Edges in COO format

        - edge_weights (torch.LongTensor): Edge weights, defined as a function of distance between atoms given by :math:`w_{i,j} = \frac{1}{d(i,j)}`, where :math:`d(i, j)` is the Euclidean distance between node :math:`i` and node :math:`j`.

        - node_pos (torch.FloatTensor): x-y-z coordinates of each node
    :rtype: Tuple
    """ 

    node_pos = torch.FloatTensor(df[['x', 'y', 'z']].to_numpy())

    kd_tree = ss.KDTree(node_pos)
    edge_tuples = list(kd_tree.query_pairs(edge_dist_cutoff))
    edges = torch.LongTensor(edge_tuples).t().contiguous()

    node_feats = torch.FloatTensor([one_of_k_encoding_unk(e, allowable_feats) for e in df[feat_col]])
    edge_weights = torch.FloatTensor(
        [1.0 / (np.linalg.norm(node_pos[i] - node_pos[j]) + 1e-5) for i, j in edge_tuples]).view(-1, 1)
    # feats = F.one_hot(elems, num_classes=len(atom_int_dict))
    
    return node_feats, edges, edge_weights.view(-1), node_pos


def mol_df_to_graph(mol, allowable_atoms=mol_atoms):
    """
    Converts molecule to a graph compatible with Pytorch-Geometric

    TODO: Change to operate on dataframe representation instead of Mol object

    :param mol: Molecule structure in RDKit format
    :type mol: rdkit.Chem.rdchem.Mol
    :param allowable_atoms: List containing allowable atom types
    :type allowable_atoms: list[str], optional

    :return: Tuple containing \n
        - node_feats (torch.FloatTensor): Features for each node, one-hot encoded by atom type in ``allowable_atoms``.
        - edges (torch.LongTensor): Edges from chemical bond graph in COO format.
        - edge_feats (torch.FloatTensor): Edge features given by bond type. Single = 1.0, Double = 2.0, Triple = 3.0, Aromatic = 1.5.
        - node_pos (torch.FloatTensor): x-y-z coordinates of each node.
    """
    node_pos = torch.FloatTensor(fo.get_coordinates_of_conformer(mol))
    bonds = fo.get_bonds_matrix_from_mol(mol)
    edge_tuples = np.argwhere(bonds)
    edges = torch.LongTensor(edge_tuples).t().contiguous()

    node_feats = torch.FloatTensor([one_of_k_encoding_unk(a.GetSymbol(), mol_atoms) for a in mol.GetAtoms()])
    edge_feats = torch.FloatTensor([bonds[i, j] for i, j in edge_tuples]).view(-1, 1)

    return node_feats, edges, edge_feats, node_pos


def combine_graphs(graph1, graph2, edges_between=True, edges_between_dist=4.5):
    """Combine two graphs into one, optionally adding edges between the two graphs using :func:`atom3d.util.graph.edges_between_graphs`. Node features are concatenated in the feature dimension, to distinguish which nodes came from which graph.

    :param graph1: One of the graphs to be combined, in the format returned by :func:`atom3d.util.graph.prot_df_to_graph` or :func:`atom3d.util.graph.mol_df_to_graph`.
    :type graph1: Tuple
    :param graph2: The other graph to be combined, in the format returned by :func:`atom3d.util.graph.prot_df_to_graph` or :func:`atom3d.util.graph.mol_df_to_graph`.
    :type graph2: Tuple
    :param edges_between: Indicates whether to add new edges between graphs, defaults to True.
    :type edges_between: bool, optional
    :param edges_between_dist: Distance cutoff in Angstroms for adding edges between graphs, defaults to 4.5.
    :type edges_between_dist: float, optional
    :return: Tuple containing \n
        - node_feats (torch.FloatTensor): Features for each node in the combined graph, concatenated along the feature dimension.\n
        - edges (torch.LongTensor): Edges of combined graph in COO format, including edges from two input graphs and edges between them, if specified.\n
        - edge_weights (torch.FloatTensor): Concatenated edge features from two input graphs and edges between them, if specified.\n
        - node_pos (torch.FloatTensor): x-y-z coordinates of each node in combined graph.
    :rtype: Tuple
    """    
    node_feats1, edges1, edge_feats1, pos1 = graph1
    node_feats2, edges2, edge_feats2, pos2 = graph2

    dummy_node_feats1 = torch.zeros(pos1.shape[0], node_feats2.shape[1])
    dummy_node_feats2 = torch.zeros(pos2.shape[0], node_feats1.shape[1])
    node_feats1 = torch.cat((node_feats1, dummy_node_feats1), dim=1)
    node_feats2 = torch.cat((dummy_node_feats2, node_feats2), dim=1)

    edges2 += pos1.shape[0]

    node_pos = torch.cat((pos1, pos2), dim=0)
    node_feats = torch.cat((node_feats1, node_feats2), dim=0)

    if edges_between:
        edges_between, edge_feats_between = edges_between_graphs(pos1, pos2)
        edge_feats = torch.cat((edge_feats1, edge_feats2, edge_feats_between), dim=0)
        edges = torch.cat((edges1, edges2, edges_between), dim=1)
    else:
        edge_feats = torch.cat((edge_feats1, edge_feats2), dim=0)
        edges = torch.cat((edges1, edges2), dim=1)

    return node_feats, edges, edge_feats, node_pos


def edges_between_graphs(pos1, pos2, dist=4.5):
    """calculates edges between nodes in two separate graphs using a specified cutoff distance.

    :param pos1: x-y-z node coordinates from Graph 1
    :type pos1: torch.FloatTensor or numpy.ndarray
    :param pos2: x-y-z node coordinates from Graph 2
    :type pos2: torch.FloatTensor or numpy.ndarray
    :return: Tuple containing\n
        - edges (torch.LongTensor): Edges between two graphs, in COO format.\n
        - edge_weights (torch.FloatTensor): Edge weights between two graphs.\n
    :rtype: Tuple
    """    
    tree1 = ss.KDTree(pos1)
    tree2 = ss.KDTree(pos2)
    res = tree1.query_ball_tree(tree2, r=dist)
    edges = []
    edge_weights = []
    for i, contacts in enumerate(res):
        if len(contacts) == 0:
            continue
        for j in contacts:
            edges.append((i, j + pos1.shape[0]))
            edge_weights.append(np.linalg.norm(pos1[i] - pos2[j]))

    edges = torch.LongTensor(edges).t().contiguous()
    edge_weights = torch.FloatTensor(edge_weights).view(-1, 1)
    return edges, edge_weights    


def adjust_graph_indices(graph):
    """Adjusts indices into graphs for concatenated multi-graph batches. Specifically, if each graph in the batch has a different selection index defined relative to that graph, the index is adjusted to be defined relative to the batch indexing.

    :param graph: Pytorch-geometric graph object representing a batch of graphs. Assumed to have a ``select_idx`` attribute set, specifying a node index for each graph
    :type graph: torch_geometric.data.Data
    :return: Same graph with selection indices adjusted
    :rtype: torch_geometric.data.Data
    """    
    batch_size = len(graph.n_nodes)
    total_n = 0
    for i in range(batch_size-1):
        n_nodes = graph.n_nodes[i].item()
        total_n += n_nodes
        graph.select_idx[i+1] += total_n
    return graph


# below functions are adapted from DeepChem repository:
def one_of_k_encoding(x, allowable_set):
    """Converts input to 1-hot encoding given a set of allowable values."""
    if x not in allowable_set:
        raise Exception("input {0} not in allowable set{1}:".format(x, allowable_set))
    return list(map(lambda s: x == s, allowable_set))


def one_of_k_encoding_unk(x, allowable_set):
    """Converts input to 1-hot encoding given a set of allowable values. Additionally maps inputs not in the allowable set to the last element."""
    if x not in allowable_set:
        x = allowable_set[-1]
    return list(map(lambda s: x == s, allowable_set))



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='atom3d',
    packages=find_packages(include=[
        'atom3d',
        'atom3d.protein',
        'atom3d.shard',
        'atom3d.util',
        'atom3d.datasets',
        'atom3d.splits',
        'atom3d.filters',
        'atom3d.data',
    ]),
    version='0.1.3',
    description='ATOM3D: Tasks On Molecules in 3 Dimensions',
    author='ATOM3D developers',
    license='MIT',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.atom3d.ai",
    install_requires=[
        'argparse',
        'biopython',
        'click',
        'easy-parallel',
        'freesasa',
        'h5py',
        'lmdb',
        'msgpack',
        'numpy',
        'pandas',
        'python-dotenv>=0.5.1',
        'scipy',
        'tables',
        'torch',
        'tqdm',
    ],
)



"""
Filtering functions for protein sequences.

"""
import pandas as pd

import atom3d.protein.sequence as seq


def form_seq_filter_against(dataset, cutoff):
    """
    Create filter that removes remove structures with too much sequence identity to a chain in supplied dataset.

    We consider each chain in each structure separately, and remove the structure if any of them matches any chain in dataset.

    :param dataset: dataset that if we are checking for matches against.
    :type dataset: atom3d dataset.
    :param cutoff: maximum allowable sequence identity to an entry in dataset before we filter.
    :type cutoff: float (0-1).

    :return: function that implements the specified filter.
    :rtype: filter function.
    """
    blast_db_path = f'blast_db'
    all_chain_sequences = [seq.get_chain_sequences(x['atoms']) for x in dataset]
    # Flatten.
    flat_chain_sequences = [x for sublist in all_chain_sequences for x in sublist]
    seq.write_to_blast_db(flat_chain_sequences, blast_db_path)

    def filter_fn(df):
        cs = seq.get_chain_sequences(df)
        hits = seq.find_similar(cs, blast_db_path, cutoff, 1)
        if len(hits) > 0:
            # Return empty dataframe.
            return df.iloc[0:0]
        else:
            return df

    return filter_fn


import atom3d.util.voxelize as vox

def graph_transform(item, structure_keys=['atoms'], label_key='labels'):
    """Transform for converting dataframes to graphs, to be applied when defining a :mod:`Dataset <atom3d.datasets.datasets>`.
    Operates on Dataset items, assumes that the item contains all keys specified in ``keys`` and ``labels`` arguments.

    :param item: Dataset item to transform
    :type item: dict
    :param keys: list of keys to transform, where each key contains a dataset of atoms, defaults to ['atoms']
    :type keys: list, optional
    :return: Transformed Dataset item
    :rtype: dict
    """    
    from torch_geometric.data import Data
    import atom3d.util.graph as gr

    for key in structure_keys:
        node_feats, edge_index, edge_feats, pos = gr.prot_df_to_graph(item[key])
        item[key] = Data(node_feats, edge_index, edge_feats, y=item[label_key], pos=pos)

    return item

def voxel_transform(item, grid_config, rot_mat=None, center_fn=vox.get_center, random_seed=None, structure_keys=['atoms']):
    """Transform for converting dataframes to voxelized grids compatible with 3D CNN, to be applied when defining a :mod:`Dataset <atom3d.datasets.datasets>`.
    Operates on Dataset items, assumes that the item contains all keys specified in ``keys`` argument.

    :param item: Dataset item to transform
    :type item: dict
    :param grid_config: Config parameters for grid. Should contain the following keys:
         `element_mapping`, dictionary mapping from element to 1-hot index; 
         `radius`, radius of grid to generate in Angstroms (half of side length); 
         `resolution`, voxel size in Angstroms;
         `num_directions`, number of directions for data augmentation (required if ``rot_mat``=None);
         `num_rolls`, number of rolls, or rotations, for data augmentation (required if ``rot_mat``=None)
    :type grid_config: :class:`dotdict <atom3d.util.voxelize.dotdict>`
    :param rot_mat: Rotation matrix (3x3) to apply to structure coordinates. If None (default), apply randomly sampled rotation according to parameters specified by ``grid_config.num_directions`` and ``grid_config.num_rolls``
    :type rot_mat: np.array
    :param center_fn: Arbitrary function for calculating the center of the voxelized grid (x,y,z coordinates) from a structure dataframe, defaults to vox.get_center
    :type center_fn: f(df -> array), optional
    :param random_seed: random seed for grid rotation, defaults to None
    :type random_seed: int, optional
    :return: Transformed Dataset item
    :rtype: dict
    """    
    for key in structure_keys:
        df = item[key]
        center = center_fn(df)

        if rot_mat is None:
            rot_mat = vox.gen_rot_matrix(grid_config, random_seed=random_seed)
        grid = vox.get_grid(
            df, center, config=grid_config, rot_mat=rot_mat)
        item[key] = grid
    return item


import torch
import numpy as np
import atom3d.datasets as da
from pathlib import Path

class MockDataset(torch.utils.data.Dataset):
    def __init__(self):
        self.data = np.arange(30)
        self.year = 2020-np.arange(30)
        self.scaf = [0]*10 + [1]*5 + [2]*3 + [3,4,5]*2 + [6,7,8,9,10,11]
        np.random.shuffle(self.scaf)
    def __len__(self):
        return len(self.data)
    def __getitem__(self,idx):
        item = {
            'data': self.data[idx],
            'year': self.year[idx],
            'scaffold': self.scaf[idx]
        }
        return item

def load_example_dataset():
    dataset = da.load_dataset(str(Path(__file__).parent.absolute()) + '/test_lmdb', 'lmdb')
    return dataset


"""Utility functions for unit vectors."""
import math

import numpy as np
import pyrr as pr


def fibonacci(num_points):
    """
    Create an approximately uniform grid on the unit sphere.

    Adapted from
    http://matplotlib.1069221.n5.nabble.com/
    Generating-N-regularly-spaced-points-on-a-sphere-td20763.html

    Args:
        num_points (int):
            the number of grid points to place.
    Returns:
        num_points x 3 numpy array where each row is x, y, and z coordinates of
        each unit vector.
    """
    inc = np.pi * (3 - np.sqrt(5))
    off = 2. / num_points
    k = np.arange(0, num_points)
    y = k * off - 1. + 0.5 * off
    r = np.sqrt(1 - y * y)
    phi = k * inc
    x = np.cos(phi) * r
    z = np.sin(phi) * r

    unit_vectors = np.zeros((num_points, 3), dtype='f4')
    unit_vectors[:, 0] = x
    unit_vectors[:, 1] = y
    unit_vectors[:, 2] = z
    return unit_vectors


def get_quaternion_alignment(uv1, up1, uv2, up2):
    """
    Return quaternion that aligns 2 pairs of direction vectors and up vectors.

    We align the (uv1, up1) to (uv2, up2).  The rotation is unique, assuming
    direction vectors and up vectors are not the same.  All vectors are assumed
    to be of unit length.  Note however that for any given rotation, two
    quaternions can represent it.

    We want:

        q * uv1 = uv2
        q * up1 = up2

    Args:
        uv1 (np.array of float):
            3 entry array specifying direction we are aligning from.
        up1 (np.array of float):
            3 entry array specifying up vector we are aligning from.
        uv2 (np.array of float):
            3 entry array specifying direction we are aligning to.
        up2 (np.array of float):
            3 entry array specifying up vector we are aligning to.

    Returns:
        pr.Quaternion specifying rotation.
    """
    uv1 = np.array(uv1)
    uv2 = np.array(uv2)
    up1 = np.array(up1)
    up2 = np.array(up2)
    # First we get quaternion to rotate uv1 to uv2.  Code from
    # http://lolengine.net/blog/2013/09/18/beautiful-maths-quaternion-from-vectors
    if np.all(np.isclose(uv1, uv2, atol=1e-5)):
        qt = pr.Quaternion()
    elif np.all(np.isclose(-uv1, uv2, atol=1e-5)):
        qt = pr.Quaternion.from_axis_rotation(up1, math.pi)
        qt.normalise()
    else:
        w = np.cross(uv1, uv2)
        qt = np.insert(w, 3, [1.0 + np.dot(uv1, uv2)])
        qt = pr.Quaternion(qt)
        qt.normalise()

    # Then, we get the quaternion to rotate the transformed up1 to up2.
    up1t = qt * pr.Vector3(up1)
    if np.all(np.isclose(up1t, up2, atol=1e-5)):
        qr = pr.Quaternion()
    elif np.all(np.isclose(-up1t, up2, atol=1e-5)):
        qr = pr.Quaternion.from_axis_rotation(uv2, math.pi)
        qr.normalise()
    else:
        w = np.cross(up1t, up2)
        qr = np.insert(w, 3, [1.0 + np.dot(up1t, up2)])
        qr = pr.Quaternion(qr)
        qr.normalise()

    # Finally, we put the two rotations together to get the final rotation.
    q = qr * qt
    return q


def rotate_v_by_q(v, q):
    """
    Rotate vector by quaternion.

    Args:
        v (np.array of float):
            3 entry array specifying vector we are rotating.
        q (pr.Quaternion):
            quaternion representing rotation we are applying.

    Returns:
        np.array of 3 floats representing rotated v.
    """
    if np.all(np.isclose(v, [0, 0, 0])):
        return v
    mag_v = np.linalg.norm(v)
    v_unit = v / mag_v
    tv_unit = q * pr.Vector3(v_unit)
    tv = tv_unit * mag_v
    return np.array(tv)


def generate_up_vectors(uv, num_rolls, up_seed=None, rev=False):
    """
    Generate evenly distributed up vectors for the suppllied direction vector.

    Args:
        uv (np.array of float):
            3 entry array specifying direction we are getting up vectors for.
        num_rolls (int):
            number of different up vectors to create.

    Keyword Args:
        up_seed (np.array of float):
            3 entry array specifying starting up direction.  If unspecified will
            be generated automatically.  Used to ensure that up vectors for two
            matching surfacelets start pointing in the same direction.
        rev (boolean):
            whether to proceed counter-clockwise as opposed to clockwise for the
            generation of the up vectors.

    Returns:
        num_rolls x 3 np.array of floats, representing the num_rolls different
        up vectors.
    """
    if up_seed is None:
        if uv[0] == 0 and uv[1] == 0:
            c = [0, -1, 0]
        else:
            c = [0, 0, -1]
        up_seed = np.cross(uv, c)
        up_seed = up_seed / np.linalg.norm(up_seed)

    if not rev:
        q = pr.Quaternion.from_axis_rotation(uv, 2 * math.pi / num_rolls)
    else:
        q = pr.Quaternion.from_axis_rotation(uv, -2 * math.pi / num_rolls)
    up = np.zeros((num_rolls, 3), dtype='f4')

    up[0] = up_seed
    for i in range(1, num_rolls):
        up[i] = rotate_v_by_q(up[i - 1], q)
        up[i] = up[i] / np.linalg.norm(up[i])
    return up


def generate_all_up_vectors(uvs, num_rolls):
    """Generate all up vectors for provided unit vectors."""
    num_uvs = uvs.shape[0]
    ups = np.zeros((num_uvs, num_rolls, 3))
    for i, uv in enumerate(uvs):
        ups[i] = generate_up_vectors(uv, num_rolls)
    return ups


def get_all_rot_mats(uvs, ups):
    """
    Generate rotation matrices for multiple unit vectors.

    TODO: Add args.
    """
    num_uvs = uvs.shape[0]
    num_ups = ups.shape[1]

    rot_mats = np.zeros((num_uvs, num_ups, 3, 3))
    for i in range(num_uvs):
        rot_mats[i] = get_rot_mats(uvs[i], ups[i])

    return rot_mats


def get_rot_mats(uv, ups):
    """
    Generate rotation matrices to transform base coordinate system to provided.

    TODO: Add docs.
    """
    num_ups = ups.shape[0]

    rot_mats = np.zeros((num_ups, 3, 3))
    for i in range(num_ups):
        rot_mats[i, :, :] = get_rot_mat(uv, ups[i])

    return rot_mats


def get_rot_mat(uv, up):
    """
    Generate rotation matrix to transforms provided coordinate system to base.

    R * uv = uv_b
    R * up = up_b
    """
    base_uv = [1., 0., 0.]
    base_up = [0., 1., 0.]

    rot_mat = np.zeros((3, 3))
    q = get_quaternion_alignment(uv, up, base_uv, base_up)
    rot_mat = np.array(pr.Matrix33(q))
    return rot_mat



"""Methods to convert between different file formats."""
import collections as col
import gzip
import os
import re

import Bio.PDB.Atom
import Bio.PDB.Chain
import Bio.PDB.Model
import Bio.PDB.Residue
import Bio.PDB.Structure
import numpy as np
import pandas as pd


# -- MANIPULATING DATAFRAMES --


def split_df(df, key):
    """
    Split dataframe containing structure(s) based on specified key. Most commonly used to split by ensemble (`key="ensemble"`) or subunit (`key=["ensemble", "subunit"]`).

    :param df: Molecular structure(s) in ATOM3D dataframe format.
    :type df: pandas.DataFrame
    :param key: key on which to split dataframe. To split on multiple keys, provide all keys in a list. Must be compatible with dataframe hierarchy, i.e. ensemble > subunit > structure > model > chain.
    :type key: Union[str, list[str]]

    :return: List of tuples containing keys and corresponding sub-dataframes.
    :rtypes: list[tuple]
    """
    return [(x, y) for x, y in df.groupby(key)]


def merge_dfs(dfs):
    """Combine a list of dataframes into a single dataframe. Assumes dataframes contain the same columns."""
    return pd.concat(dfs).reset_index(drop=True)


# -- CONVERTING INTERNAL FORMATS --


def bp_to_df(bp):
    """Convert biopython representation to ATOM3D dataframe representation.

    :param bp: Molecular structure in Biopython representation.
    :type bp: Bio.PDB.Structure

    :return: Molecular structure in ATOM3D dataframe format.
    :rtype: pandas.DataFrame
    """
    df = col.defaultdict(list)
    for atom in Bio.PDB.Selection.unfold_entities(bp, 'A'):
        residue = atom.get_parent()
        chain = residue.get_parent()
        model = chain.get_parent()
        df['ensemble'].append(bp.get_id())
        df['subunit'].append(0)
        df['structure'].append(bp.get_id())
        df['model'].append(model.serial_num)
        df['chain'].append(chain.id)
        df['hetero'].append(residue.id[0])
        df['insertion_code'].append(residue.id[2])
        df['residue'].append(residue.id[1])
        df['segid'].append(residue.segid)
        df['resname'].append(residue.resname)
        df['altloc'].append(atom.altloc)
        df['occupancy'].append(atom.occupancy)
        df['bfactor'].append(atom.bfactor)
        df['x'].append(atom.coord[0])
        df['y'].append(atom.coord[1])
        df['z'].append(atom.coord[2])
        df['element'].append(atom.element)
        df['name'].append(atom.name)
        df['fullname'].append(atom.fullname)
        df['serial_number'].append(atom.serial_number)
    df = pd.DataFrame(df)
    return df


def df_to_bp(df_in):
    """Convert ATOM3D dataframe representation to biopython representation. Assumes dataframe contains only one structure.

    :param df_in: Molecular structure in ATOM3D dataframe format.
    :type df_in: pandas.DataFrame

    :return: Molecular structure in BioPython format.
    :rtype: Bio.PDB.Structure
    """
    all_structures = df_to_bps(df_in)
    if len(all_structures) > 1:
        raise RuntimeError('More than one structure in provided dataframe.')
    return all_structures[0]


def df_to_bps(df_in):
    """Convert ATOM3D dataframe representation containing multiple structures to list of Biopython structures. Assumes different structures are specified by `ensemble` and `structure` columns of dataframe.

    :param df_in: Molecular structures in ATOM3D dataframe format.
    :type df_in: pandas.DataFrame

    :return : List of molecular structures in BioPython format.
    :rtype: list[Bio.PDB.Structure]
    """
    df = df_in.copy()
    all_structures = []
    for (structure, s_atoms) in split_df(df_in, ['ensemble', 'structure']):
        new_structure = Bio.PDB.Structure.Structure(structure[1])
        for (model, m_atoms) in df.groupby(['model']):
            new_model = Bio.PDB.Model.Model(model)
            for (chain, c_atoms) in m_atoms.groupby(['chain']):
                new_chain = Bio.PDB.Chain.Chain(chain)
                for (residue, r_atoms) in c_atoms.groupby(
                        ['hetero', 'residue', 'insertion_code']):
                    # Take first atom as representative for residue values.
                    rep = r_atoms.iloc[0]
                    new_residue = Bio.PDB.Residue.Residue(
                        (rep['hetero'], rep['residue'], rep['insertion_code']),
                        rep['resname'], rep['segid'])
                    for row, atom in r_atoms.iterrows():
                        new_atom = Bio.PDB.Atom.Atom(
                            atom['name'],
                            [atom['x'], atom['y'], atom['z']],
                            atom['bfactor'],
                            atom['occupancy'],
                            atom['altloc'],
                            atom['fullname'],
                            atom['serial_number'],
                            atom['element'])
                        new_residue.add(new_atom)
                    new_chain.add(new_residue)
                new_model.add(new_chain)
            new_structure.add(new_model)
        all_structures.append(new_structure)
    return all_structures


# -- READING FILES -- #


## general reader function to get a Biopython structure
#  (not suported: sharded, silent, xyz-gdb)


def read_any(f, name=None):
    """Read any ATOM3D file type into Biopython structure (compatible with pdb, pdb.gz, mmcif, sdf, xyz).

    :param f: file path
    :type f: Union[str, Path]
    :param name: optional name or identifier for structure. If None (default), use file basename.
    :type name: str

    :return: Biopython object containing structure
    :rtype: Bio.PDB.Structure
    """
    if is_pdb(f):
        return read_pdb(f, name)
    elif is_pdb_gz(f):
        return read_pdb_gz(f, name)
    elif is_mmcif(f):
        return read_mmcif(f, name)
    elif is_sdf(f):
        return read_sdf(f)
    elif is_xyz(f):
        return read_xyz(f, name)
    else:
        raise ValueError(f"Unrecognized filetype for {f:}")


## functions to check file format


patterns = {
    'pdb': r'pdb[0-9]*$',
    'pdb.gz': r'pdb[0-9]*\.gz$',
    'mmcif': r'(mm)?cif$',
    'sdf': r'sdf[0-9]*$',
    'xyz': r'xyz[0-9]*$',
    'xyz-gdb': r'xyz[0-9]*$',
    'silent': r'out$',
    'sharded': r'@[0-9]+',
}

_regexes = {k: re.compile(v) for k, v in patterns.items()}


def is_pdb(f):
    """Check if file is in pdb format."""
    return _regexes['pdb'].search(str(f))


def is_pdb_gz(f):
    """Check if file is in mmcif format."""
    return _regexes['pdb.gz'].search(str(f))


def is_mmcif(f):
    """Check if file is in mmcif format."""
    return _regexes['mmcif'].search(str(f))


def is_sdf(f):
    """Check if file is in sdf format."""
    return _regexes['sdf'].search(str(f))


def is_xyz(f):
    """Check if file is in xyz format."""
    return _regexes['xyz'].search(str(f))


def is_sharded(f):
    """Check if file is in sharded format."""
    return _regexes['sharded'].search(str(f))


## reader functions for specific file formats


def read_pdb(pdb_file, name=None):
    """Read pdb file into Biopython structure.

    :param pdb_file: file path
    :type pdb_file: Union[str, Path]
    :param name: optional name or identifier for structure. If None (default), use file basename.
    :type name: str

    :return: Biopython object containing structure
    :rtype: Bio.PDB.Structure
    """
    if name is None:
        name = os.path.basename(pdb_file)
    parser = Bio.PDB.PDBParser(QUIET=True)
    bp = parser.get_structure(name, pdb_file)
    return bp


def read_pdb_gz(pdb_gz_file, name=None):
    """Read pdb.gz file into Biopython structure.

    :param pdb_gz_file: file path
    :type pdb_gz_file: Union[str, Path]
    :param name: optional name or identifier for structure. If None (default), use file basename.
    :type name: str

    :return: Biopython object containing structure
    :rtype: Bio.PDB.Structure
    """
    if name is None:
        name = os.path.basename(pdb_gz_file)
    parser = Bio.PDB.PDBParser(QUIET=True)
    bp = parser.get_structure(
        name, gzip.open(pdb_gz_file, mode='rt', encoding='latin1'))
    return bp


def read_mmcif(mmcif_file, name=None):
    """Read mmCIF file into Biopython structure.

    :param mmcif_file: file path
    :type mmcif_file: Union[str, Path]
    :param name: optional name or identifier for structure. If None (default), use file basename.
    :type name: str

    :return: Biopython object containing structure
    :rtype: Bio.PDB.Structure
    """
    if name is None:
        os.path.basename(mmcif_file)
    parser = Bio.PDB.MMCIFParser(QUIET=True)
    return parser.get_structure(name, mmcif_file)


def read_sdf(sdf_file, name=None, sanitize=False, add_hs=False, remove_hs=False):
    """Read SDF file into Biopython structure.

    :param sdf_file: file path
    :type sdf_file: Union[str, Path]
    :param sanitize: sanitize structure with RDKit.
    :type sanitize: bool
    :param add_hs: add hydrogen atoms with RDKit.
    :type add_hs: bool
    :param remove_hs: remove hydrogen atoms with RDKit.
    :type remove_hs: bool

    :return: Biopython object containing structure
    :rtype: Bio.PDB.Structure
    """

    dflist = []
    molecules = read_sdf_to_mol(sdf_file, sanitize=sanitize,
                                add_hs=add_hs, remove_hs=remove_hs)
    for im,m in enumerate(molecules):
        if m is not None:
            df = mol_to_df(m, residue=im,
                           ensemble=m.GetProp("_Name"),
                           structure=m.GetProp("_Name"),
                           model=m.GetProp("_Name"))
            dflist.append(df)
    assert len(dflist) >= 1
    if len(dflist) > 1:
        bp = df_to_bp(merge_dfs(dflist))
    else:
        bp = df_to_bp(dflist[0])

    return bp


def read_sdf_to_mol(sdf_file, sanitize=False, add_hs=False, remove_hs=False):
    """Reads a list of molecules from an SDF file.

    :param add_hs: Specifies whether to add hydrogens. Defaults to False
    :type add_hs: bool
    :param remove_hs: Specifies whether to remove hydrogens. Defaults to False
    :type remove_hs: bool
    :param sanitize: Specifies whether to sanitize the molecule. Defaults to False
    :type sanitize: bool

    :return: list of molecules in RDKit format.
    :rtype: list[rdkit.Chem.rdchem.Mol]
    """
    from rdkit import Chem

    suppl = Chem.SDMolSupplier(sdf_file, sanitize=sanitize, removeHs=remove_hs)

    molecules = [mol for mol in suppl]

    if add_hs:
        for mol in molecules:
            if mol is not None:
                mol = Chem.AddHs(mol, addCoords=True)

    return molecules


def mol_to_df(mol, add_hs=False, structure=None, model=None, ensemble=None, residue=999):
    """
    Convert molecule in RDKit format to ATOM3D dataframe format, with PDB-style columns.

    :param mol: Molecule in RDKit format.
    :type mol: rdkit.Chem.rdchem.Mol

    :return: Dataframe in standard ATOM3D format.
    :rtype: pandas.DataFrame
    """

    from rdkit import Chem
    df = col.defaultdict(list)
    if add_hs:
        mol = Chem.AddHs(mol, addCoords=True)
    conf = mol.GetConformer()
    for i, a in enumerate(mol.GetAtoms()):
        position = conf.GetAtomPosition(i)
        df['ensemble'].append(ensemble)
        df['structure'].append(structure)
        df['model'].append(model)
        df['chain'].append('LIG')
        df['hetero'].append('')
        df['insertion_code'].append('')
        df['residue'].append(residue)
        df['segid'].append('')
        df['resname'].append('LIG')
        df['altloc'].append('')
        df['occupancy'].append(1)
        df['bfactor'].append(0)
        df['x'].append(position.x)
        df['y'].append(position.y)
        df['z'].append(position.z)
        df['element'].append(a.GetSymbol().upper())
        df['serial_number'].append(i)
    df = pd.DataFrame(df)
    # Make up atom names
    elements = df['element'].unique()
    el_count = {}
    for e in elements:
        el_count[e] = 0
    new_name = []
    for el in df['element']:
        el_count[el] += 1
        new_name.append('%s%i'%(el,el_count[el]))
    df['name'] = new_name
    df['fullname'] = new_name
    return df


def read_xyz(xyz_file, name=None, gdb=False):
    """Read an XYZ file into Biopython representation (optionally with GDB9-specific data)

    :param inputfile: Path to input file in XYZ format.
    :type inputfile: Union[str, Path]
    :param gdb_data: Specifies whether to process and return GDB9-specific data.
    :type gdb_date: bool

    :return: If `gdb=False`, returns Biopython Structure object containing molecule structure. If `gdb=True`, returns tuple containing \n
        \t- bp (Bio.PDB.Structure): Biopython object containing molecule structure.\n
        \t- data (list[float]): Scalar molecular properties.\n
        \t- freq (list[float]): Harmonic vibrational frequencies (:math:`3n_{atoms}−5` or :math:`3n_{atoms}-6`, in :math:`cm^{−1}`).\n
        \t- smiles (str): SMILES string from GDB-17 and from B3LYP relaxation.\n
        \t- inchi (str): InChI string for Corina and B3LYP geometries.
    """
    # Load the xyz file into a dataframe
    if gdb:
        df, data, freq, smiles, inchi = read_xyz_to_df(xyz_file, gdb_data=True)
    else:
        df = read_xyz_to_df(xyz_file)
    if name is not None: df.index.name = name
    # Make up atom names
    elements = df['element'].unique()
    el_count = {}
    for e in elements:
        el_count[e] = 0
    new_name = []
    for el in df['element']:
        el_count[el] += 1
        new_name.append('%s%i'%(el,el_count[el]))
    # Fill additional fields
    df['ensemble'] = [df.name.replace(' ','_')]*len(df)
    df['subunit'] = [0]*len(df)
    df['structure'] = [df.name.replace(' ','_')]*len(df)
    df['model'] = [0]*len(df)
    df['chain'] = ['L']*len(df)
    df['hetero'] = ['']*len(df)
    df['insertion_code'] = ['']*len(df)
    df['residue'] = [1]*len(df)
    df['segid'] = ['LIG']*len(df)
    df['resname'] = ['LIG']*len(df)
    df['altloc'] = ['']*len(df)
    df['occupancy'] = [1.]*len(df)
    df['bfactor'] = [0.]*len(df)
    df['name'] = new_name
    df['fullname'] = new_name
    df['serial_number'] = range(len(df))
    # Convert to biopython representation
    bp = df_to_bp(df)
    if gdb:
        return bp, data, freq, smiles, inchi
    else:
        return bp


def read_xyz_to_df(inputfile, gdb_data=False):
    """Read an XYZ file into Pandas DataFrame representation (optionally with GDB9-specific data)

    :param inputfile: Path to input file in XYZ format.
    :type inputfile: Union[str, Path]
    :param gdb_data: Specifies whether to process and return GDB9-specific data.
    :type gdb_date: bool

    :return: If `gdb=False`, returns DataFrame containing molecule structure. If `gdb=True`, returns tuple containing\n
        \t- molecule (pandas.DataFrame): Pandas DataFrame containing molecule structure.\n
        \t- data (list[float]): Scalar molecular properties. Returned only when `gdb=True`.\n
        \t- freq (list[float]): Harmonic vibrational frequencies (:math:`3n_{atoms}−5` or :math:`3n_{atoms}-6`, in :math:`cm^{−1}`).  Returned only when `gdb=True`.\n
        \t- smiles (str): SMILES string from GDB-17 and from B3LYP relaxation. Returned only when `gdb=True`.\n
        \t- inchi (str): InChI string for Corina and B3LYP geometries. Returned only when `gdb=True`.\n
    """
    with open(inputfile) as f:
        # Reading number of atoms in the molecule
        num_atoms = int(f.readline().strip())
        # Loading GDB ID and label data
        line_labels = f.readline().strip().split('\t')
        name = line_labels[0]
        if gdb_data: data = [float(ll) for ll in line_labels[1:]]
        # Skip atom data (will be read using pandas below)
        for n in range(num_atoms): f.readline()
        # Harmonic vibrational frequencies
        if gdb_data:
            freq = [float(ll) for ll in f.readline().strip().split('\t')]
        # SMILES and InChI
        if gdb_data: smiles = f.readline().strip().split('\t')[0]
        if gdb_data: inchi  = f.readline().strip().split('\t')[0]
    # Define columns: element, x, y, z, Mulliken charges (GDB only)
    columns = ['element','x', 'y', 'z']
    if gdb_data: columns += ['charge']
    # Load atom information
    molecule = pd.read_table(inputfile, names=columns,
                             skiprows=2, nrows=num_atoms,
                             delim_whitespace=True)
    # Name the dataframe
    molecule.name = name
    molecule.index.name = name
    # return molecule info
    if gdb_data:
        return molecule, data, freq, smiles, inchi
    else:
        return molecule


# -- WRITING FILES --


def write_pdb(out_file, structure, **kwargs):
    """Write a biopython structure to a pdb file. This function accepts any viable arguments to Bio.PDB.PDBIO.save() as keyword arguments.

    :param out_file: Path to output PDB file.
    :type out_file: Union[str, Path]
    :param structure: Biopython object containing protein structure.
    :type structure: Bio.PDB.Structure
    """
    io = Bio.PDB.PDBIO()
    io.set_structure(structure)
    io.save(out_file, **kwargs)
    return


def write_mmcif(out_file, structure, **kwargs):
    """Write a biopython structure to an mmcif file. This function accepts any viable arguments to Bio.PDB.MMCIFIO.save() as keyword arguments.

    :param out_file: Path to output mmCIF file.
    :type out_file: Union[str, Path]
    :param structure: Biopython object containing protein structure.
    :type structure: Bio.PDB.structure
    """
    io = Bio.PDB.MMCIFIO()
    io.set_structure(structure)
    io.save(out_file)
    return


# -- CONVENIENCE FUNCTIONS AND CONSTANTS--
#    (for custom data conversions)

atomic_number = {'H': 1, 'He': 2, 'Li': 3, 'Be': 4, 'B': 5, 'C': 6, 'N': 7, 'O': 8, 'F': 9, 'Ne': 10,
                 'Na': 11, 'Mg': 12, 'Al': 13, 'Si': 14, 'P': 15, 'S': 16, 'Cl': 17, 'Ar': 18, 'K': 19, 'Ca': 20,
                 'Sc': 21, 'Ti': 22, 'V': 23, 'Cr': 24, 'Mn': 25, 'Fe': 26, 'Co': 27, 'Ni': 28, 'Cu': 29, 'Zn': 30,
                 'Ga': 31, 'Ge': 32, 'As': 33, 'Se': 34, 'Br': 35, 'Kr': 36, 'Rb': 37, 'Sr': 38, 'Y': 39, 'Zr': 40,
                 'Nb': 41, 'Mo': 42, 'Tc': 43, 'Ru': 44, 'Rh': 45, 'Pd': 46, 'Ag': 47, 'Cd': 48, 'In': 49, 'Sn': 50,
                 'Sb': 51, 'Te': 52, 'I': 53, 'Xe': 54, 'Cs': 55, 'Ba': 56, 'La': 57, 'Ce': 58, 'Pr': 59, 'Nd': 60,
                 'Pm': 61, 'Sm': 62, 'Eu': 63, 'Gd': 64, 'Tb': 65, 'Dy': 66, 'Ho': 67, 'Er': 68, 'Tm': 69, 'Yb': 70,
                 'Lu': 71, 'Hf': 72, 'Ta': 73, 'W': 74, 'Re': 75, 'Os': 76, 'Ir': 77, 'Pt': 78, 'Au': 79, 'Hg': 80,
                 'Tl': 81, 'Pb': 82, 'Bi': 83, 'Po': 84, 'At': 85, 'Rn': 86, 'Fr': 87, 'Ra': 88, 'Ac': 89, 'Th': 90,
                 'Pa': 91, 'U': 92, 'Np': 93, 'Pu': 94, 'Am': 95, 'Cm': 96, 'Bk': 97, 'Cf': 98, 'Es': 99, 'Fm': 100,
                 'Md': 101, 'No': 102, 'Lr': 103, 'Rf': 104, 'Db': 105, 'Sg': 106, 'Bh': 107, 'Hs': 108, 'Mt': 109, 'Ds': 110,
                 'Rg': 111, 'Cn': 112, 'Nh': 113, 'Fl': 114, 'Mc': 115, 'Lv': 116, 'Ts': 117, 'Og': 118}


def get_coordinates_from_df(df):
    """Extract XYZ coordinates from molecule in dataframe format.

    :param df: Dataframe containing molecular structure. Must have columns named `x`, `y`, and `z`.
    :type df: pandas.DataFrame

    :return: XYZ coordinates as N x 3 array
    :rtype: numpy.ndarray
    """
    xyz = np.empty([len(df), 3])

    xyz[:, 0] = np.array(df.x)
    xyz[:, 1] = np.array(df.y)
    xyz[:, 2] = np.array(df.z)

    return xyz


def get_coordinates_of_conformer(mol):
    """Reads the coordinates of a conformer.

    :params mol: Molecule in RDKit format.
    :type mol: rdkit.Chem.rdchem.Mol

    :return: XYZ coordinates of molecule as N x 3 float array.
    :rtype: numpy.ndarray
    """

    symb = [a.GetSymbol() for a in mol.GetAtoms()]
    conf = mol.GetConformer()
    xyz = np.empty([mol.GetNumAtoms(), 3])

    for ia, name in enumerate(symb):
        position = conf.GetAtomPosition(ia)
        xyz[ia] = np.array([position.x, position.y, position.z])

    return xyz


def get_connectivity_matrix_from_mol(mol):
    """Calculates the binary bond connectivity matrix from a molecule.

    :param mol: Molecule in RDKit format.
    :type mol: rdkit.Chem.rdchem.Mol

    :return: Binary connectivity matrix (N x N) containing all molecular bonds.
    :rtype: numpy.ndarray
    """

    # Initialization
    num_at = mol.GetNumAtoms()
    connect_matrix = np.zeros([num_at, num_at], dtype=int)

    # Go through all atom pairs and check for bonds between them
    for a in mol.GetAtoms():
        for b in mol.GetAtoms():
            bond = mol.GetBondBetweenAtoms(a.GetIdx(), b.GetIdx())
            if bond is not None:
                connect_matrix[a.GetIdx(), b.GetIdx()] = 1

    return connect_matrix


def get_bonds_matrix_from_mol(mol):
    """
    Calculates matrix of bond types from a molecule and returns as numpy array.
    Bond types are encoded as double:
     single bond (1.0)
     double bond (2.0)
     triple bond (3.0)
     aromatic bond (1.5).

    :param mol: Molecule in RDKit format.
    :type mol: rdkit.Chem.rdchem.Mol

    :return: Bond matrix (N x N) with bond types encoded as double.
    :rtype: numpy.ndarray

    """

    # Initialization
    num_at = mol.GetNumAtoms()
    bonds_matrix = np.zeros([num_at, num_at])

    # Go through all atom pairs and check for bonds between them
    for a in mol.GetAtoms():
        for b in mol.GetAtoms():
            bond = mol.GetBondBetweenAtoms(a.GetIdx(), b.GetIdx())
            if bond is not None:
                bt = bond.GetBondTypeAsDouble()
                bonds_matrix[a.GetIdx(), b.GetIdx()] = bt

    return bonds_matrix


def get_bonds_list_from_mol(mol):
    """
    Calculates all bonds and bond types from a molecule and returns as dataframe.
    Bond types are encoded as double:
     single bond (1.0)
     double bond (2.0)
     triple bond (3.0)
     aromatic bond (1.5).

    :param mol: Molecule in RDKit format.
    :type mol: rdkit.Chem.rdchem.Mol

    :return: Bond information as dataframe with columns `atom1`, `atom2`, `type`.
    :rtype: pandas.DataFrame

    """
    bonds_list = []
    for b in mol.GetBonds():
        atom1 = b.GetBeginAtomIdx()
        atom2 = b.GetEndAtomIdx()
        btype = b.GetBondTypeAsDouble()
        bonds_list.append([atom1,atom2,btype])
    col = ['atom1','atom2','type']
    bonds_df = pd.DataFrame(bonds_list, columns=col)
    return bonds_df



"""Functions related to SCOP classes of structures."""
import os

import pandas as pd


SCOP_CLA_LATEST_FILE = 'atom3d/data/metadata/scop-cla-latest.txt'
PDB_CHAIN_SCOP2_UNIPROT_FILE = 'atom3d/data/metadata/pdb_chain_scop2_uniprot.csv'
PDB_CHAIN_SCOP2B_SF_UNIPROT_FILE = 'atom3d/data/metadata/pdb_chain_scop2b_sf_uniprot.csv'


def get_scop_index():
    """Get index mapping from PDB code and chain to SCOP classification."""

    # Load core SCOP database.  Mapping from domains to classification.
    scop = pd.read_csv(
        SCOP_CLA_LATEST_FILE, skiprows=6, delimiter=' ',
        names=['fa-domid', 'fa-pdbid', 'fa-pdbreg', 'fa-uniid', 'fa-unireg',
               'sf-domid', 'sf-pdbid', 'sf-pdbreg', 'sf-uniid', 'sf-unireg',
               'scop'])
    scop['pdb_code'] = scop['fa-pdbid'].apply(lambda x: x.lower())
    scop['type'] = \
        scop['scop'].apply(lambda x: int(x.split(',')[0].split('=')[1]))
    scop['class'] = \
        scop['scop'].apply(lambda x: int(x.split(',')[1].split('=')[1]))
    scop['fold'] =  \
        scop['scop'].apply(lambda x: int(x.split(',')[2].split('=')[1]))
    scop['superfamily'] = \
        scop['scop'].apply(lambda x: int(x.split(',')[3].split('=')[1]))
    scop['family'] = \
        scop['scop'].apply(lambda x: int(x.split(',')[4].split('=')[1]))
    del scop['scop']

    # Load mapping of representatives to scop domains.
    scop2_uniprot = pd.read_csv(
        PDB_CHAIN_SCOP2_UNIPROT_FILE, skiprows=2,
        names=['pdb_code', 'chain', 'sp-primary', 'sf-domid', 'fa-domid'],
    )
    # Some superfamily entries are a bit messed up here, so we remove.
    scop2_uniprot = scop2_uniprot[scop2_uniprot['sf-domid'].str.isnumeric()]
    scop2_uniprot['sf-domid'] = pd.to_numeric(scop2_uniprot['sf-domid'])

    # Load mapping of non-representatives to scop domains.
    scop2b_sf_uniprot = pd.read_csv(
        PDB_CHAIN_SCOP2B_SF_UNIPROT_FILE, skiprows=2, usecols=[0, 1, 2, 3],
        names=['pdb_code', 'chain', 'sf-domid', 'sp-primary'])

    # Now merge the mappings:
    # (pdb_code, chain) -> domain
    # AND
    # domain -> scop class
    # TO GET
    # (pdb_code, chain) -> scop class
    scop = scop.set_index('sf-domid')
    scop2_uniprot = scop2_uniprot.set_index('sf-domid')
    scop2b_sf_uniprot = scop2b_sf_uniprot.set_index('sf-domid')
    sf_to_scop = scop[['class', 'fold', 'superfamily', 'family']]
    sf_to_chain = pd.concat([
        scop2_uniprot[['pdb_code', 'chain']],
        scop2b_sf_uniprot[['pdb_code', 'chain']]])
    chain_to_scop = pd.merge(sf_to_scop, sf_to_chain, on='sf-domid')
    chain_to_scop = chain_to_scop.drop_duplicates().set_index(
        ['pdb_code', 'chain']).sort_index(level=[0, 1])
    return chain_to_scop



"""
Filtering functions for protein data bank files.

These all are applied to individual atom dataframes, and remove entries from that dataframe as necessary.
"""
import pandas as pd

PDB_ENTRY_TYPE_FILE = 'atom3d/data/metadata/pdb_entry_type.txt'
RESOLUTION_FILE = 'atom3d/data/metadata/resolu.idx'


def form_size_filter(max_size=None, min_size=None):
    """
    Create filter for a certain number of residues, keeping only structures that fall within the specified bounds.

    :param max_size: maximum allowable number of residues in a structure.  None means there is no maximum.
    :type max_size: int.
    :param min_size: minimum allowable number of residues in a structure.  None means there is no minimum.
    :type min_size: int.

    :return: function that implements the specified filter.
    :rtype: filter function.
    """

    def filter_fn(df):
        to_keep = pd.Series([True] * len(df), index=df['structure'])
        residues = \
            df[['structure', 'model', 'chain', 'residue']].drop_duplicates()
        counts = residues['structure'].value_counts()
        if max_size:
            tmp = counts <= max_size
            tmp = tmp[df['structure']]
            to_keep = (to_keep & tmp)
        if min_size:
            tmp = counts >= min_size
            tmp = tmp[df['structure']]
            to_keep = (to_keep & tmp)
        return df[to_keep.values]
    return filter_fn


def form_source_filter(allowed=None, excluded=None):
    """
    Create filter for experimental source.

    Valid entries are: diffraction, NMR, EM, other.  Only one of allowed and excluded can be set.

    :param allowed: allowed experimental sources.
    :type allowed: list[str].
    :param excluded: excluded experimental sources.
    :type excluded: list[str].

    :return: function that implements the specified filter.
    :rtype: filter function.
    """
    if excluded is not None and allowed is not None:
        raise RuntimeError('Can only specify one of allowed and excluded.')
    if excluded is None:
        excluded = []
    if allowed is None:
        allowed = []
    pdb_entry_type = pd.read_csv(PDB_ENTRY_TYPE_FILE, delimiter='\t',
                                 names=['pdb_code', 'molecule_type', 'source'])
    pdb_entry_type['pdb_code'] = \
        pdb_entry_type['pdb_code'].apply(lambda x: x.lower())
    pdb_entry_type = pdb_entry_type.set_index('pdb_code', drop=True)
    source = pdb_entry_type['source']

    def filter_fn(df):
        pdb_codes = df['structure'].apply(lambda x: x[:4].lower())

        if len(allowed) > 0:
            to_keep = source[pdb_codes].isin(allowed)
        elif len(excluded) > 0:
            to_keep = ~source[pdb_codes].isin(excluded)
        else:
            to_keep = pd.Series([True] * len(df), index=df['structure'])
        return df[to_keep.values]
    return filter_fn


def form_molecule_type_filter(allowed=None, excluded=None):
    """
    Create filter for molecule type.

    Valid entries are: prot, prot-nuc, nuc, carb, other.  Only one of allowed and excluded can be set.

    :param allowed: allowed molecule types.
    :type allowed: list[str].
    :param excluded: excluded molecule types.
    :type excluded: list[str].

    :return: function that implements the specified filter.
    :rtype: filter function.
    """
    if allowed is None:
        allowed = []
    if excluded is None:
        excluded = []
    pdb_entry_type = pd.read_csv(PDB_ENTRY_TYPE_FILE, delimiter='\t',
                                 names=['pdb_code', 'molecule_type', 'source'])
    pdb_entry_type['pdb_code'] = \
        pdb_entry_type['pdb_code'].apply(lambda x: x.lower())
    pdb_entry_type = pdb_entry_type.set_index('pdb_code')
    molecule_type = pdb_entry_type['molecule_type']

    def filter_fn(df):
        pdb_codes = df['structure'].apply(lambda x: x[:4].lower())

        if len(allowed) > 0:
            to_keep = molecule_type[pdb_codes].isin(allowed)
        elif len(excluded) > 0:
            to_keep = ~molecule_type[pdb_codes].isin(excluded)
        else:
            to_keep = pd.Series([True] * len(df), index=df['structure'])
        return df[to_keep.values]
    return filter_fn


def form_resolution_filter(threshold):
    """
    Create filter for experimental resolution.

    :param threshold: maximum allowable experimental resolution.
    :type threshold: float.

    :return: function that implements the specified filter.
    :rtype: filter function.
    """
    resolution = pd.read_csv(RESOLUTION_FILE, skiprows=6, delimiter='\t',
                             usecols=[0, 2], names=['pdb_code', 'resolution'])
    resolution['pdb_code'] = resolution['pdb_code'].apply(lambda x: x.lower())
    resolution = resolution.set_index('pdb_code')['resolution']

    # Remove duplicates beyond the first.
    resolution = resolution[~resolution.index.duplicated()]

    def filter_fn(df):
        pdb_codes = df['structure'].apply(lambda x: x[:4].lower())
        to_keep = resolution[pdb_codes] < threshold
        return df[to_keep.values]
    return filter_fn
